from dataclasses import dataclass

from nwn.types import GenderedLanguage
from nwn.gff._impl import FieldKind


class Byte(int):
    FIELD_KIND = FieldKind.BYTE
    SIMPLE_DATA_FORMAT = "B"
    LABEL = "byte"

    def __new__(cls, value):
        if not 0 <= value <= 255:
            raise ValueError(f"BYTE value out of bounds: {value}")
        return super().__new__(cls, value)


class Char(int):
    FIELD_KIND = FieldKind.CHAR
    SIMPLE_DATA_FORMAT = "b"
    LABEL = "char"

    def __new__(cls, value):
        if not -128 <= value <= 127:
            raise ValueError(f"CHAR value out of bounds: {value}")
        return super().__new__(cls, value)


class Word(int):
    FIELD_KIND = FieldKind.WORD
    SIMPLE_DATA_FORMAT = "H"
    LABEL = "word"

    def __new__(cls, value):
        if not 0 <= value <= 65535:
            raise ValueError(f"WORD value out of bounds: {value}")
        return super().__new__(cls, value)


class Short(int):
    FIELD_KIND = FieldKind.SHORT
    SIMPLE_DATA_FORMAT = "h"
    LABEL = "short"

    def __new__(cls, value):
        if not -32768 <= value <= 32767:
            raise ValueError(f"SHORT value out of bounds: {value}")
        return super().__new__(cls, value)


class Dword(int):
    FIELD_KIND = FieldKind.DWORD
    SIMPLE_DATA_FORMAT = "I"
    LABEL = "dword"

    def __new__(cls, value):
        if not 0 <= value <= 4294967295:
            raise ValueError(f"DWORD value out of bounds: {value}")
        return super().__new__(cls, value)


class Int(int):
    FIELD_KIND = FieldKind.INT
    SIMPLE_DATA_FORMAT = "i"
    LABEL = "int"

    def __new__(cls, value):
        if not -2147483648 <= value <= 2147483647:
            raise ValueError(f"INT value out of bounds: {value}")
        return super().__new__(cls, value)


class Dword64(int):
    FIELD_KIND = FieldKind.DWORD64
    LABEL = "dword64"

    def __new__(cls, value):
        if not 0 <= value <= 18446744073709551615:
            raise ValueError(f"DWORD64 value out of bounds: {value}")
        return super().__new__(cls, value)


class Int64(int):
    FIELD_KIND = FieldKind.INT64
    LABEL = "int64"

    def __new__(cls, value):
        if not -9223372036854775808 <= value <= 9223372036854775807:
            raise ValueError(f"INT64 value out of bounds: {value}")
        return super().__new__(cls, value)


class Float(float):
    FIELD_KIND = FieldKind.FLOAT
    SIMPLE_DATA_FORMAT = "f"
    LABEL = "float"


class Double(float):
    FIELD_KIND = FieldKind.DOUBLE
    LABEL = "double"


class CExoString(str):
    FIELD_KIND = FieldKind.CEXOSTRING
    LABEL = "cexostring"


class ResRef(str):
    FIELD_KIND = FieldKind.RESREF
    LABEL = "resref"

    def __new__(cls, value):
        if len(value) > 16:
            raise ValueError(f"ResRef value too long: {value}")
        return super().__new__(cls, value)


@dataclass
class CExoLocString:
    """Represents a localized string in the NWN engine."""

    FIELD_KIND = FieldKind.CEXOLOCSTRING
    LABEL = "cexolocstring"

    strref: Dword
    entries: dict[GenderedLanguage, str]

    def __str__(self):
        english = GenderedLanguage.from_id(0)
        if self.strref == 0xFFFFFFFF and self.entries.keys() == {english}:
            return self.entries[english]
        s = {str(lid.to_id()): text for lid, text in self.entries.items()}
        if self.strref != 0xFFFFFFFF:
            s["id"] = str(self.strref)
        return str(s)


class VOID(bytes):
    FIELD_KIND = FieldKind.VOID
    LABEL = "void"

    def __new__(cls, value):
        return super().__new__(cls, value)


class Struct(dict):
    """GFF Structs are just python dicts with .attr access and some metadata."""

    FIELD_KIND = FieldKind.STRUCT
    LABEL = "struct"

    def __init__(self, struct_id, **kwargs):
        super().__init__(**kwargs)
        object.__setattr__(self, "_struct_id", struct_id)

    @property
    def struct_id(self):
        """The struct ID of the struct."""
        return object.__getattribute__(self, "_struct_id")

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as exc:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{item}'"
            ) from exc

    def __setattr__(self, name, value):
        self[name] = value


class List(list[Struct]):
    """
    GFF Lists are just python lists of Structs. They carry no metadata.

    This class exists as a convenience for type checking and
    future extensibility.
    """

    FIELD_KIND = FieldKind.LIST
    LABEL = "list"

    def __init__(self, value=None):
        if value is None:
            value = []
        super().__init__(value)


SIMPLE_TYPES = {
    FieldKind.BYTE: Byte,
    FieldKind.CHAR: Char,
    FieldKind.WORD: Word,
    FieldKind.SHORT: Short,
    FieldKind.DWORD: Dword,
    FieldKind.INT: Int,
    FieldKind.FLOAT: Float,
}

LABEL_MAP = {
    "byte": Byte,
    "char": Char,
    "word": Word,
    "short": Short,
    "dword": Dword,
    "int": Int,
    "dword64": Dword64,
    "int64": Int64,
    "float": Float,
    "double": Double,
    "cexostring": CExoString,
    "resref": ResRef,
    "cexolocstring": CExoLocString,
    "void": VOID,
    "struct": Struct,
    "list": List,
}


def type_label_to_type(label: str):
    """
    Convert a GFF type label to the corresponding type class.

    Args:
        label: The GFF label to convert.

    Returns:
        The corresponding type class.

    Raises:
        ValueError: If the label is not recognized.
    """

    if label in LABEL_MAP:
        return LABEL_MAP[label]
    raise ValueError(f"Unknown GFF type label: {label}")



"""
Class and library to deal with NWN resources directories.
"""

import os
from pathlib import Path

from nwn.res import Container, is_valid_resref


class LocalDirectory(Container):
    """
    A resource directory that reads files from a filesystem directory.

    Data is indexed once on initialization; you need to explicitly call
    reindex() to refresh the file list if you change the contents of the
    directory outside of the LocalDirectory instance.

    Missing/non-existent directories are allowed for reading.

    If you initialize with writable=True, the directory will be created if it
    does not exist.

    This class implements a read-write mapping interface but if you intend
    to modify the directory contents, you must initialize it with writable=True.

    Args:
        path: The filesystem path to the resource directory.
    """

    def reindex(self):
        self._files = (
            {
                f.name.lower(): f
                for f in self._path.iterdir()
                if f.is_file() and is_valid_resref(f.name)
            }
            if self._path.is_dir()
            else {}
        )

    def __init__(self, path: str | Path, writable: bool = False):
        self._path = Path(path)
        self.reindex()
        self._writable = writable
        if self._writable:
            self._path.mkdir(parents=True, exist_ok=True)

    def __getitem__(self, key: str) -> bytes:
        file_path = self._files[key.lower()]
        with open(file_path, "rb") as f:
            return f.read()

    def __iter__(self):
        return iter(self._files)

    def __len__(self) -> int:
        return len(self._files)

    def __setitem__(self, key: str, value: bytes):
        if not self._writable:
            raise TypeError("ResDir is read-only")
        if not is_valid_resref(key):
            raise ValueError(f"Invalid resref: {key}")
        file_path = self._path / key
        with open(file_path, "wb") as f:
            f.write(value)
        self._files[key.lower()] = file_path

    def __delitem__(self, key: str):
        if not self._writable:
            raise TypeError("ResDir is read-only")
        file_path = self._files.pop(key.lower())
        os.remove(file_path)

    def __repr__(self):
        return f"LocalDirectory({self._path}, count={len(self)})"



from dataclasses import dataclass, field
from enum import Enum

from nwn.types import Vector


class _ObjectSelf:
    """Use Object.SELF."""

    def __repr__(self):
        return "Object.SELF"


@dataclass(frozen=True)
class Object:
    """
    A Object is a immutable reference to a game object in the world.
    This maps directly to the nwscript ``object`` type.
    """

    INVALID = None
    """Special value: Invalid is 0x7F000000 in nwscript."""

    SELF = _ObjectSelf()
    """
    Special value: SELF is a placeholder for the calling object in a script.

    This has no meaning outside of a script context and is not a real reference.
    """

    id: int = 0x7F000000

    def __post_init__(self):
        if self.id < 0 or self.id > 0x7FFFFFFF:
            raise ValueError(f"Invalid object id: {self.id}")

    def __bool__(self):
        """
        Objects are truthy if they're not INVALID, even if the reference
        itself is dangling in game logic.
        """
        return self.id != 0x7F000000

    def __eq__(self, other):
        if other is None:
            return self is Object.INVALID
        if isinstance(other, Object):
            return self.id == other.id
        raise TypeError(f"Cannot compare Object to {type(other)}")

    def __repr__(self):
        if self is Object.INVALID:
            return "Object.INVALID"
        return f"Object(0x{self.id:X})"


Object.INVALID = Object(0x7F000000)


@dataclass
class Location:
    """
    Maps directly to the nwscript ``location`` type.
    """

    INVALID = None

    position: Vector = field(default_factory=Vector)
    orientation: float = 0.0
    area: Object = field(default_factory=Object.INVALID)


@dataclass
class Effect:
    """
    Maps directly to the NWScript effect type.

    Not fully implemented: This is just a stub for now.
    """

    INVALID = None

    type: int = 0


Effect.INVALID = Effect(type=0)


@dataclass(frozen=True)
class EngineStructure:
    """
    Placeholder type for engst until specific types are implemented.
    """

    vm_type: "VMType"


@dataclass(frozen=True)
class _VMClosure:
    """
    A closure as stored by the VM.
    """

    ip: int
    sp: int
    bp: int


class VMType(Enum):
    """
    All types supported by the NWScript VM as implemented in EE 37.
    """

    VOID = "void"
    INT = "int"
    FLOAT = "float"
    STRING = "string"
    OBJECT = "object"
    VECTOR = "vector"
    LOCATION = "location"
    EFFECT = "effect"
    ITEMPROPERTY = "itemproperty"
    EVENT = "event"
    TALENT = "talent"
    SQLQUERY = "sqlquery"
    CASSOWARY = "cassowary"
    JSON = "json"
    ACTION = "action"

    def python_type(self):
        """
        Returns the python type that the VM uses for this VMType on the VM stack.

        Returns:
            type: The python class/type.
        """

        match self:
            case VMType.INT:
                return int
            case VMType.FLOAT:
                return float
            case VMType.STRING:
                return str
            case VMType.OBJECT:
                return Object
            case VMType.VECTOR:
                return Vector
            case VMType.LOCATION:
                return Location
            case VMType.EFFECT:
                return Effect
            case VMType.ACTION:
                return _VMClosure
            case (
                VMType.ITEMPROPERTY,
                VMType.EVENT,
                VMType.TALENT,
                VMType.SQLQUERY,
                VMType.CASSOWARY,
                VMType.JSON,
            ):
                return EngineStructure(self)
            case _:
                raise NotImplementedError(f"Unknown VMType: {self}")



"""A parser for .set files (tilesets configuration)."""

import configparser
from dataclasses import dataclass, field
import re
from typing import TextIO, get_origin, get_args


@dataclass
class Terrain:
    name: str = ""
    strref: int = 0


@dataclass
class Crosser:
    name: str = ""
    strref: int = 0


@dataclass
class Rule:
    placed: str
    placedheight: int
    adjacent: str
    adjacentheight: int
    changed: str
    changedheight: int


@dataclass
class Door:
    type: int
    x: float = 0
    y: float = 0
    z: float = 0
    orientation: float = 0


@dataclass
class Tile:
    doors: list[Door]

    model: str
    walkmesh: str = ""
    topleft: str = ""
    topleftheight: int = 0
    topright: str = ""
    toprightheight: int = 0
    bottomleft: str = ""
    bottomleftheight: int = 0
    bottomright: str = ""
    bottomrightheight: int = 0
    top: str = ""
    right: str = ""
    bottom: str = ""
    left: str = ""
    mainlight1: int = 0
    mainlight2: int = 0
    sourcelight1: int = 0
    sourcelight2: int = 0
    animloop1: int = 0
    animloop2: int = 0
    animloop3: int = 0
    sounds: int = 0
    pathnode: str = ""
    orientation: float = 0
    imagemap2d: str = ""


@dataclass
class Group:
    name: str
    rows: int
    columns: int
    tiles: list[int] = field(default_factory=list)


@dataclass
class SetGrass:
    grass: int
    density: float | None = None
    height: float | None = None
    ambientred: float | None = None
    ambientgreen: float | None = None
    ambientblue: float | None = None
    diffusered: float | None = None
    diffusegreen: float | None = None
    diffuseblue: float | None = None


@dataclass
class Set:
    name: str
    type: str
    version: str
    interior: int
    hasheighttransition: int
    envmap: str
    transition: float
    displayname: int
    border: str
    default: str
    floor: str

    grass: SetGrass | None = None
    terrains: list[Terrain] = field(default_factory=list)
    crossers: list[Crosser] = field(default_factory=list)
    primary_rules: list[Rule] = field(default_factory=list)
    tiles: list[Tile] = field(default_factory=list)
    groups: list[Group] = field(default_factory=list)


def _dataclass_name(s):
    return "".join(x or "_" for x in s.split("_"))


def _read_value(ty, v):
    # Handle optional unions (e.g., float | None)
    if get_origin(ty) is not None:
        args = get_args(ty)
        if len(args) == 2 and type(None) in args:
            non_none_type = next(arg for arg in args if arg is not type(None))
            return _read_value(non_none_type, v) if v is not None else None

    if ty == int:
        return int(v) if v else 0
    if ty == bool:
        return v in {"1"}
    if ty == float:
        return float(v) if v else 0.0
    if ty == str:
        return str(v)
    if get_origin(ty) == list or ty == list:
        # We manually load these in
        return []
    raise ValueError(f"Unsupported type: {ty}")


def _read_dataclass(cls, **kwargs):
    return cls(
        **{
            field.name: _read_value(field.type, kwargs[_dataclass_name(field.name)])
            for field in cls.__dataclass_fields__.values()
            if _dataclass_name(field.name) in kwargs and not field.name.startswith("_")
        }
    )


def read_set(file: TextIO) -> Set:
    """
    Reads a .set file and parses its contents into a Set object.
    """

    general = None

    data = configparser.ConfigParser()
    data.read_file(file)

    for s in data.sections():
        if s == "GENERAL":
            general = _read_dataclass(Set, **data[s])

        if not general:
            raise ValueError("No GENERAL section found")

        if s == "GRASS":
            grass = _read_dataclass(SetGrass, **data[s])
            general.grass = grass

        if ma := re.match(r"^TERRAIN(\d+)$", s):
            tid = int(ma.group(1))
            terrain = _read_dataclass(Terrain, **data[s])
            general.terrains.append(terrain)

        if ma := re.match(r"^CROSSER(\d+)$", s):
            terrain = _read_dataclass(Crosser, **data[s])
            general.crossers.append(terrain)

        if ma := re.match(r"^PRIMARY RULE(\d+)$", s):
            rule = _read_dataclass(Rule, **data[s])
            general.primary_rules.append(rule)

        if ma := re.match(r"^TILE(\d+)$", s):
            tid = int(ma.group(1))
            tile = _read_dataclass(Tile, **data[s])
            general.tiles.append(tile)

        if ma := re.match(r"^TILE(\d+)DOOR(\d+)$", s):
            tid = int(ma.group(1))
            door = _read_dataclass(Door, **data[s])
            tile = general.tiles[tid]
            tile.doors.append(door)

        if ma := re.match(r"^GROUP(\d+)$", s):
            group = _read_dataclass(Group, **data[s])
            count = group.rows * group.columns
            for i in range(count):
                tile = data[s][f"Tile{i}"]
                group.tiles.append(int(tile))
            general.groups.append(group)

    if not general:
        raise ValueError("No GENERAL section found; invalid tileset?")

    return general



"""
Read and write TLK (Talk Table) files (for translation and base game string references).

All strings are transparently converted to and from the NWN encoding.
"""

import struct
from typing import BinaryIO

from nwn.types import Language
from nwn.environ import get_codepage


class Entry(str):
    """
    A single entry in a TLK file including the text, sound resref, and sound length.
    Only used when reading TLK files with sound data.
    """

    def __new__(cls, text: str, sound_resref: str = "", sound_length: float = 0.0):
        obj = str.__new__(cls, text)
        object.__setattr__(obj, "sound_resref", sound_resref)
        object.__setattr__(obj, "sound_length", sound_length)
        return obj

    sound_resref: str = ""
    sound_length: float = 0.0

    @property
    def text(self) -> str:
        return str(self)


def read(file: BinaryIO, max_entries=0x7FFFF) -> tuple[list[Entry], Language]:
    """
    Reads a TLK file fully into memory and returns a list of entries.

    Args:
        file: A binary file object containing the TLK file.
        max_entries: The maximum number of entries to read from the TLK file.
            This is a sanity check to avoid allocating excess memory when
            reading untrusted or corrupted data.

    Returns:
        A tuple containing the list of entries, and the language of the TLK file.

    Raises:
        ValueError: If the file does not contain valid TLK data.
    """

    (
        magic,
        version,
        language,
        entry_count,
        entries_offset,
    ) = struct.unpack("<4s4sIII", file.read(20))

    if magic != b"TLK ":
        raise ValueError("Invalid TLK magic")
    if version != b"V3.0":
        raise ValueError("Invalid TLK version")
    language = Language(language)

    if entry_count > max_entries:
        raise ValueError(f"Too many entries in TLK file: {entry_count} > {max_entries}")

    file.seek(20)
    entry_data = file.read(entry_count * 40)
    entries = [
        (
            sound_resref.decode("ascii").strip("\x00\xc0"),
            offset_to_string,
            string_sz,
            sound_length,
        )
        for (
            _,  # flags (text_present=0x1,sound_present=0x2,length_present=0x4)
            sound_resref,
            _,  # volume_variance unused as per spec
            _,  # pitch_variance unused as per spec
            offset_to_string,
            string_sz,
            sound_length,
        ) in struct.iter_unpack("<I16sIIIIf", entry_data)
    ]

    # We might be reading from a file-like object in the middle of some other stream,
    # so make sure not to read beyond the boundary of the last string.
    end_offset = max(offset + size for (_, offset, size, _) in entries)
    file.seek(entries_offset)
    string_data = file.read(end_offset).decode(get_codepage())

    return [
        Entry(string_data[offset : offset + size], sound_resref, sound_length)
        for (sound_resref, offset, size, sound_length) in entries
    ], language


def write(file: BinaryIO, entries: list[Entry], language: Language):
    """
    Writes a Tlk object to a binary file.

    Args:
        file: A binary file object to write the TLK data to.
        entries: A list containing the entries to write, in order.
            Entries can be either strings or Entry objects.
        language: The language of the TLK file to write.

    Raises:
        ValueError: If the TLK object contains invalid data.
    """

    file.write(
        struct.pack(
            "<4s4sIII",
            b"TLK ",
            b"V3.0",
            language.value,
            len(entries),
            20 + len(entries) * 40,
        )
    )

    str_data = bytearray()

    string_offset = 0
    for idx, entry in enumerate(entries):
        flags = 0
        text_len = 0

        if entry.text:
            flags |= 0x1
            text_len = len(entry.text.encode(get_codepage()))
        if entry.sound_resref:
            flags |= 0x2
        if entry.sound_length:
            flags |= 0x4

        if len(entry.sound_resref) > 16:
            raise ValueError(f"Sound resref at {idx} is too long")

        file.write(
            struct.pack(
                "<I16sIIIIf",
                flags,
                entry.sound_resref.ljust(16, "\x00").encode("ascii"),
                0,  # volume variance: unused as per spec
                0,  # pitch variance: unused as per spec
                string_offset,
                text_len,
                entry.sound_length,
            )
        )
        string_offset += text_len
        str_data += entry.text.encode(get_codepage())

    file.write(str_data)



"""
Read and write compressedbuf files, used in NWSync and campaign databases.

Mainly used for NWSync file storage; but also used in sqlite files such as
nwsync storage and string/blob compression in campaign sqlite databases.

Binary Format
=============

+-------------------+-------------------+--------------------------------------+
| Field             | Size (bytes)      | Description                          |
+===================+===================+======================================+
| Magic             | 4                 | `FileMagic`                          |
+-------------------+-------------------+--------------------------------------+
| Header Version    | 4                 | Format version (currently always 3)  |
+-------------------+-------------------+--------------------------------------+
| Algorithm         | 4                 | `Algorithm`                          |
+-------------------+-------------------+--------------------------------------+
| Uncompressed Size | 4                 | Size of the uncompressed data        |
+-------------------+-------------------+--------------------------------------+
| Algorithm Version | 4 (if ZLIB/ZSTD)  | Version of the compression algorithm |
|                   |                   | (currently always 1)                 |
+-------------------+-------------------+--------------------------------------+
| Dictionary        | 4 (if ZSTD)       | Dictionary ID (currently always 0)   |
+-------------------+-------------------+--------------------------------------+
| Compressed Data   | Variable          | The compressed data                  |
+-------------------+-------------------+--------------------------------------+
"""

from enum import IntEnum
from typing import BinaryIO
from io import BytesIO
import zlib
import pyzstd as zstd

from .types import FileMagic


class Algorithm(IntEnum):
    NONE = 0
    ZLIB = 1
    """Deprecated, use ZSTD instead."""
    ZSTD = 2


_VERSION = 3
_ZLIB_VERSION = 1
_ZSTD_VERSION = 1


def read(
    file: BinaryIO, expect_magic: FileMagic | None = None
) -> tuple[bytes, FileMagic, Algorithm]:
    """
    Decompresses the given binary file using the specified algorithm.

    Args:
        file: The binary file to decompress.
        expect_magic: The expected file magic. If provided,
            the file magic read from the file must match this value.

    Returns:
        A tuple containing the decompressed data, the file magic, and the algorithm used.

    Raises:
        ValueError: If the stream is invalid, or the magic number does not
            match the expected value, or if an unsupported algorithm is encountered.
    """

    magic = FileMagic(file.read(4))
    if expect_magic is not None and magic != expect_magic:
        raise ValueError(f"invalid magic: {magic}")

    header_version = int.from_bytes(file.read(4), "little")
    if header_version != _VERSION:
        raise ValueError(f"invalid header version: {header_version}")

    algorithm = Algorithm(int.from_bytes(file.read(4), "little"))
    uncompressed_size = int.from_bytes(file.read(4), "little")

    if uncompressed_size == 0:
        return (b"", magic, algorithm)

    if algorithm == Algorithm.NONE:
        return (file.read(uncompressed_size), magic, algorithm)

    elif algorithm == Algorithm.ZLIB:
        vers = int.from_bytes(file.read(4), "little")
        if vers != _ZLIB_VERSION:
            raise ValueError(f"invalid zlib header version: {vers}")
        return (zlib.decompress(file.read()), magic, algorithm)

    elif algorithm == Algorithm.ZSTD:
        vers = int.from_bytes(file.read(4), "little")
        if vers != _ZSTD_VERSION:
            raise ValueError(f"invalid zstd header version: {vers}")
        dictionary = int.from_bytes(file.read(4), "little")
        if dictionary != 0:
            raise ValueError("dictionaries are not supported")

        return (zstd.decompress(file.read()), magic, algorithm)

    else:
        raise NotImplementedError()


def write(file: BinaryIO, magic: FileMagic, data: bytes, alg=Algorithm.ZSTD):
    """
    Compresses the given data and writes it to the specified file using the specified algorithm.

    Args:
        file: The file object to write the compressed data to.
        magic: The file magic to write.
        data: The data to be compressed.
        alg (Algorithm, optional): The compression algorithm to use. Defaults to Algorithm.ZSTD.

    Raises:
        ValueError: If an unsupported algorithm is specified.
    """

    file.write(magic)
    file.write(_VERSION.to_bytes(4, "little"))
    file.write(alg.to_bytes(4, "little"))
    file.write(len(data).to_bytes(4, "little"))

    if alg == Algorithm.NONE:
        file.write(data)

    elif alg == Algorithm.ZLIB:
        file.write(_ZLIB_VERSION.to_bytes(4, "little"))
        file.write(zlib.compress(data))

    elif alg == Algorithm.ZSTD:
        file.write(_ZSTD_VERSION.to_bytes(4, "little"))
        file.write((0).to_bytes(4, "little"))  # dictionary
        file.write(zstd.compress(data))

    else:
        raise ValueError("Unsupported algorithm")


def compress(data: bytes, file_magic: FileMagic, alg=Algorithm.ZSTD) -> bytes:
    """
    Compresses the given data using the specified algorithm and returns the
    compressed data as bytes.

    Args:
        data: The data to be compressed.
        file_magic: The file magic to use.
        alg (Algorithm, optional): The compression algorithm to use.
            Defaults to Algorithm.ZSTD.

    Returns:
        The compressed data as bytes.
    """

    out = BytesIO()
    write(out, file_magic, data, alg)
    return out.getvalue()


def decompress(data: bytes, expect_magic: FileMagic | None = None) -> bytes:
    """
    Decompresses the given data using the specified algorithm and returns
    the decompressed data as bytes.

    Args:
        data: The data to be decompressed.
        expect_magic: The expected file magic. If provided,
            the file magic read from the data must match this value.

    Returns:
        The decompressed data as bytes.
    """
    inp = BytesIO(data)
    decompressed_data, _, _ = read(inp, expect_magic)
    return decompressed_data



"""
nwn
===

A python package with various modules for accessing Neverwinter Nights: Enhanced Edition
data formats and functionality.

Stability
---------

This package is currently in ALPHA state. API stability is not guaranteed.

Installation
------------

The package is available on PyPI and can be installed with pip:

.. code-block:: bash

    pip install nwn

License
-------

This package is licensed under the MIT license.
"""

# Deprecation import shim, will be removed at some point; these are just left
# here to avoid breaking existing code that imports from nwn.*
from .environ import get_codepage as get_nwn_encoding
from .types import FileMagic, Gender, GenderedLanguage, Language
from .res import restype_to_extension, extension_to_restype, is_valid_resref



from typing import Any

from nwn.types import FileMagic, GenderedLanguage
from nwn.gff import (
    Struct,
    List,
    Byte,
    Char,
    Word,
    Short,
    Dword,
    Int,
    Dword64,
    Int64,
    Float,
    Double,
    CExoString,
    ResRef,
    CExoLocString,
    VOID,
)

_TYPE_MAP = {
    "byte": Byte,
    "char": Char,
    "word": Word,
    "short": Short,
    "dword": Dword,
    "int": Int,
    "dword64": Dword64,
    "int64": Int64,
    "float": Float,
    "double": Double,
    "cexostring": CExoString,
    "resref": ResRef,
    "cexolocstring": CExoLocString,
    "void": VOID,
}


def _value_from_json(data: dict) -> Any:
    type_name = data["type"]
    value = data["value"]
    if type_name == "list":
        return List([_struct_from_json(item, 0) for item in value])
    if type_name == "struct":
        return _struct_from_json(value, 0)
    if type_name == "cexolocstring":
        strref = value.get("id", 0xFFFFFFFF)
        entries = {
            GenderedLanguage.from_id(int(k)): v for k, v in value.items() if k != "id"
        }
        return CExoLocString(strref, entries)
    if type_name in _TYPE_MAP:
        gff_type = _TYPE_MAP[type_name]
        return gff_type(value)
    raise ValueError(f"Unknown type: {type_name}")


def _struct_from_json(data: dict, default_struct_id: int) -> Struct:
    struct_id = data.get("__struct_id", default_struct_id)
    fields = {}
    for key, value in data.items():
        if not key.startswith("__"):
            fields[key] = _value_from_json(value)
    return Struct(struct_id, **fields)


def struct_from_json(data: dict) -> tuple[Struct, FileMagic]:
    """
    Parse nwn-style json representation into native python types, same
    as you would get from reading a GFF file with the `read` function.

    The data provided must be a dictionary of the root structure; e.g. reading
    the file as written by the game with the json module.

    Example:
        >>> with open("item.uti.json", "r") as f:
        ...     data = json.load(f)
        ...     print(data["AddCost"])  # "AddCost": {"type": "dword", "value": 1000}
        ...     from nwn.gff import struct_from_json
        ...     xfrm, ty = struct_from_json(data)
        ...     print(xfrm["AddCost"])  # Dword(1000)

    Args:
        data: The json data representing a GFF structure, as emitted by
            the game or neverwinter.nim.

    Returns:
        The parsed GFF structure.

    Raises:
        ValueError: If the json data is malformed or contains unsupported types.
    """
    return (_struct_from_json(data, 0xFFFFFFFF), FileMagic(data["__data_type"]))


def _value_to_json(value: Any) -> dict:
    type_name = value.FIELD_KIND.name.lower()
    if isinstance(value, CExoLocString):
        entries: dict[str, Any] = {str(k.to_id()): v for k, v in value.entries.items()}
        if value.strref and value.strref != 0xFFFFFFFF:
            entries["id"] = value.strref
        return {
            "type": "cexolocstring",
            "value": entries,
        }
    if isinstance(value, (str, int, float)):
        return {"type": type_name, "value": value}
    raise ValueError(f"Unsupported value type: {type(value)} for value: {value}")


def _struct_to_json(struct: Struct, data_type: FileMagic | None = None) -> dict:
    result = {"__struct_id": struct.struct_id}
    if data_type:
        result["__data_type"] = data_type.decode("utf-8")
    for key, value in struct.items():
        if isinstance(value, Struct):
            result[key] = {"type": "struct", "value": _struct_to_json(value)}
        elif isinstance(value, list):
            result[key] = {
                "type": "list",
                "value": [_struct_to_json(item) for item in value],
            }
        elif hasattr(value, "FIELD_KIND"):
            result[key] = _value_to_json(value)
        else:
            raise ValueError(f"Unsupported value type: {type(value)} for key: {key}")
    return result


def struct_to_json(struct: Struct, data_type: FileMagic) -> dict:
    """
    Convert a Struct to its nwn-style json representation.

    Args:
        struct: The Struct to convert.
        data_type: File magic type to include in the json.

    Returns:
        A dictionary representing the Struct in json format.

    Raises:
        ValueError: If the Struct contains unsupported types.
    """
    return _struct_to_json(struct, data_type)



"""
Bindings for the official nwn script compiler.

The compiler depends on a native library. This package ships with precompiled
binaries for MacOS, Windows and Linux both on X64 and ARM64 architectures.

This module or the underlying library is currently NOT threadsafe.

Example:
    >>> from nwn.nwscript.comp import Compiler
    ...
    ... def resolver(filename: str) -> bytes | None:
    ...     # Load the file from disk, or return None if not found.
    ...     # You can also service from memory, or anywhere else.
    ...     with open(filename, "rb") as f:
    ...         return f.read()
    ... comp = Compiler(resolver)
    >>> ncs, ndb = comp.compile("myscript.nss")
    >>> print(ncs)  # The compiled bytecode
    >>> print(ndb)  # The debug information

"""

import os
import platform
import ctypes as ct
from ctypes.util import find_library
from typing import Callable
from enum import IntEnum

from nwn.res import restype_to_extension
from nwn.environ import get_codepage


_CB_WRITE = ct.CFUNCTYPE(
    ct.c_int32,
    ct.c_char_p,
    ct.c_uint16,
    ct.c_void_p,
    ct.c_size_t,
    ct.c_bool,
)

_CB_LOAD = ct.CFUNCTYPE(
    ct.c_char_p,
    ct.c_char_p,
    ct.c_uint16,
)


class Optimization(IntEnum):
    DEAD_FUNCTIONS = 0x00000001
    MELD_INSTRUCTIONS = 0x00000002
    DEAD_BRANCHES = 0x00000004


class _NativeCompileResult(ct.Structure):
    _fields_ = [
        ("code", ct.c_int32),
        ("str", ct.c_char_p),
    ]


class CompilationError(Exception):
    """Exception raised for errors in the compilation process."""

    def __init__(self, code: int, message: str):
        assert code != 0
        super().__init__()
        self.code = code
        self.message = message

    def __str__(self):
        return f"{self.message} ({self.code})"


class Compiler:
    """A class to compile NWScript using the NWScript compiler."""

    _instance: ct.CDLL | None = None

    def __init__(
        self,
        resolver: Callable[[str], str | bytes | None],
        src_rt=2009,  # nss
        bin_rt=2010,  # ncs
        dbg_rt=2064,  # ndb
        langspec=b"nwscript",
        debug_info=True,
        max_include_depth=16,
        encoding=None,
        optimizations: set[Optimization] | int | None = None,
    ):
        """
        Create a new compiler instance.

        This will load the compiler library, and immediately request the langspec
        file (e.g. nwscript.nss).

        Re-use the compiler instance for increased performance.

        Args:
            resolver: A callable that resolves the filename to a str, or None.
                The filename will always have the source restype extension appended.
                (e.g. "myscript.nss", not "myscript").
            src_rt: The resource type for the source file.
            bin_rt: The resource type for the binary file.
            dbg_rt: The resource type for the debug file.
            langspec: The language specification.
            debug_info: Whether to write debug information.
                Generating NDB is usually cheap and useful for debugging, so
                the recommendation is to generate and store this info alongside
                the NCS file.
            max_include_depth: The maximum include depth.
            encoding: The encoding to use for filenames (defaults to the configured
                nwn encoding).
            optimizations: A set of optimizations to apply. The default is
                to use whatever the linked library defaults to.

        Raises:
            FileNotFoundError: If the library cannot not be found, either bundled
                or in the system library path.
            OSError: If the library could not be loaded.
        """

        self._resolver = resolver
        self._ncs = None
        self._ndb = None
        self._comp = None
        self._src_rt = src_rt
        self._bin_rt = bin_rt
        self._dbg_rt = dbg_rt
        self._src_ext = restype_to_extension(src_rt)
        self._bin_ext = restype_to_extension(bin_rt)
        self._dbg_ext = restype_to_extension(dbg_rt)
        self._encoding = encoding or get_codepage()

        self._lib = Compiler._load_library()

        def cb_write(*args):
            return self._write_file(*args)

        def cb_load(*args):
            return self._load_file(*args)

        self._cb_write = _CB_WRITE(cb_write)
        self._cb_load = _CB_LOAD(cb_load)
        self._comp = self._lib.scriptCompApiNewCompiler(
            src_rt,
            bin_rt,
            dbg_rt,
            self._cb_write,
            self._cb_load,
        )

        # NB: This will trigger the first request, for "nwscript.nss".
        self._lib.scriptCompApiInitCompiler(
            self._comp,
            langspec,
            debug_info,
            max_include_depth,
            None,
            b"scriptout",
        )

        if optimizations is not None:
            if isinstance(optimizations, int):
                opt = optimizations
            elif isinstance(optimizations, set):
                opt = sum(o.value for o in optimizations)
            else:
                raise TypeError(f"Invalid optimizations type: {type(optimizations)}")

            self._lib.scriptCompApiSetOptimizationFlags(self._comp, opt)

    @classmethod
    def _load_library(cls) -> ct.CDLL:
        if cls._instance:
            return cls._instance

        sys = platform.system().lower()
        if sys == "darwin":
            sys = "macos"
        arch = platform.machine().lower()
        if arch == "arm64":
            arch = "aarch64"
        if arch == "amd64":
            arch = "x86_64"
        ext = "so" if sys == "linux" else "dll" if sys == "windows" else "dylib"

        root_fn = "nwnscriptcomp"
        lib_path = os.path.join(
            os.path.dirname(__file__),
            "lib",
            f"{sys}_{arch}",
            f"lib{root_fn}.{ext}",
        )

        if not lib_path or not os.path.exists(lib_path):
            lib_path = find_library(root_fn)

        if not lib_path or not os.path.exists(lib_path):
            raise FileNotFoundError(
                "Could not find libscriptcomp (no builtin, no system)"
            )

        cls._instance = ct.cdll.LoadLibrary(lib_path)

        if not cls._instance or not cls._instance._name:
            raise OSError("Could not load libscriptcomp (failed to instance)")

        lib = cls._instance

        abiver = lib.scriptCompApiGetABIVersion()
        if abiver != 1:
            raise ImportError(f"ABI version mismatch: {abiver} != 1")

        fn_newcomp = lib.scriptCompApiNewCompiler
        fn_newcomp.argtypes = [
            ct.c_int,  # src
            ct.c_int,  # bin
            ct.c_int,  # dbg
            _CB_WRITE,
            _CB_LOAD,
        ]
        fn_newcomp.restype = ct.c_void_p

        fn_initcomp = lib.scriptCompApiInitCompiler
        fn_initcomp.argtypes = [
            ct.c_void_p,  # compiler
            ct.c_char_p,  # lang
            ct.c_bool,  # writeDebug
            ct.c_int,  # maxIncludeDepth
            ct.c_char_p,  # graphvizOut
            ct.c_char_p,  # outputAlias
        ]

        fn_compile = lib.scriptCompApiCompileFile
        fn_compile.argtypes = [
            ct.c_void_p,  # compiler
            ct.c_char_p,  # filename
        ]
        fn_compile.restype = _NativeCompileResult

        fn_deliver_file = lib.scriptCompApiDeliverFile
        fn_deliver_file.argtypes = [
            ct.c_void_p,  # compiler
            ct.c_void_p,  # data
            ct.c_size_t,  # size
        ]

        fn_destroycomp = lib.scriptCompApiDestroyCompiler
        fn_destroycomp.argtypes = [
            ct.c_void_p,  # compiler
        ]

        fn_setopt = lib.scriptCompApiSetOptimizationFlags
        fn_setopt.argtypes = [
            ct.c_void_p,  # compiler
            ct.c_int,  # flags
        ]

        return lib

    def __del__(self):
        if self._comp:
            self._lib.scriptCompApiDestroyCompiler(self._comp)

    def _write_file(self, _fn, res_type: int, data: bytes, size: int, is_binary: bool):
        dat = bytes(ct.cast(data, ct.POINTER(ct.c_char * size)).contents)
        if is_binary:
            assert not self._ncs, f"compiler called twice for binary data: {self._ncs=}"
            assert res_type == self._bin_rt, f"unexpected restype: {res_type=}"
            self._ncs = dat
        else:
            assert not self._ndb, f"compiler called twice for debug data: {self._ndb=}"
            assert res_type == self._dbg_rt, f"unexpected restype: {res_type=}"
            self._ndb = dat.decode(self._encoding)
        return 0

    def _load_file(self, script_name: bytes, res_type: int):
        assert res_type == self._src_rt, f"unexpected restype: {res_type=}"
        filename = f"{script_name.decode(self._encoding)}.{self._src_ext}"
        if data := self._resolver(filename):
            data = data.encode(self._encoding) if isinstance(data, str) else data
            self._lib.scriptCompApiDeliverFile(self._comp, data, len(data))
            return True

        return False

    def compile(self, script_name: str) -> tuple[bytes, str]:
        """
        Compile the given script filename.

        This will invoke multiple callbacks to load the script and
        any includes. The script will be compiled to bytecode and
        debug information.

        Args:
            script_name: The script to compile. It will be requested
                from the resolver. The source restype extension is
                optionally stripped; other extensions will raise
                a ValueError.

        Returns:
            A tuple of (bytecode, debug data).
            The bytecode is a bytes object, and the debug data is
            a string; or None if no debug data was requested.

        Raises:
            CompilationError: If the compilation fails.
            ValueError: If the script name is invalid.
        """

        root, ext = os.path.splitext(script_name)

        if ext == "." + self._src_ext:
            script_name = root
        elif ext:
            raise ValueError(
                f"Invalid script name: {script_name}. Expected extension: {self._src_ext}"
            )

        r = self._lib.scriptCompApiCompileFile(
            self._comp, script_name.encode(self._encoding)
        )
        if r.code != 0:
            raise CompilationError(r.code, r.str.decode(self._encoding).strip())
        assert self._ncs
        ncs, ndb = self._ncs, self._ndb
        self._ncs = None
        self._ndb = None
        return (ncs, ndb)



import logging

from nwn.nwscript import ndb
from nwn.nwscript.asm import Opcode, Auxcode, read_instr
from nwn.nwscript.langspec import LanguageSpec
from nwn.nwscript import Object, VMType
from nwn.types import Vector

from .._types import _VMClosure
from ._stack import Stack
from ._script import Script

_logger = logging.getLogger(__name__)


class _ScriptExecutionEnd(Exception):
    pass


class VM:
    """
    VM Script Executor for running NWScript bytecode.

    Supports the following type mappings:

    - int <> int (32 bit on the nwscript side)
    - float <> float (32 bit on the nwscript side)
    - str <> string: in the current encoding (get_codepage())
    - Object <> object
    - Vector <> vector
    - Location <> location
    - custom structs <> dict: unset keys will push the type default value; \
        extra keys are ignored.
    - Engine structure support is very preliminary and incomplete
    """

    def __init__(self, script: Script, spec: LanguageSpec, impl):
        """
        Create a new VM.

        You need one VM per script. You can reuse spec and impl between them.

        The implementation must be a class that has the methods that correspond to
        the VM command names. The methods must accept the same number of arguments
        as the VM command expects, mapped to the corresponding python types.

        After creating the VM, you can call the `run` method to execute the
        script, or use call() or proxy() to call NWScript functions directly.

        Args:
            script: The script to run.
            spec: The language spec your script was compiled against.
            impl: The implementation of the script.
        """
        self._spec = spec
        self._impl = impl

        self._script = script
        self._ncs = script.ncs
        self._ndb = script.ndb

        self._stack = Stack()
        self._ret: list = []
        self._object_self = Object.INVALID

        # TODO: these should probably also move into stack.py
        self._bp = 0
        self._save_ip = 0
        self._save_bp = 0
        self._save_sp = 0

        # Super janky, doesn't really work yet.
        self._stop_before_main = False

    @property
    def object_self(self):
        """
        Return the object ID that the script is running on. Defaults to INVALID.
        """
        return self._object_self

    @object_self.setter
    def object_self(self, v):
        """
        Set/change the object ID that the script is running on.

        Changing this will affect the OBJECT_SELF constant in the script.
        """
        self._object_self = v

    @property
    def impl(self):
        """
        Return the implementation object that this VM was created with.
        """
        return self._impl

    @property
    def spec(self):
        """
        Return the language spec that this VM was created with.
        """
        return self._spec

    @property
    def sp(self):
        """
        Get the current stack pointer.
        """
        return self._stack.sp

    @sp.setter
    def sp(self, v):
        """
        Set the current stack pointer. Currently used internally.
        """
        self._stack.sp = v

    @property
    def ip(self):
        """
        Return the current instruction pointer (offset into bytecode).
        """
        return self._ncs.tell()

    @ip.setter
    def ip(self, v):
        """
        Manually set the instruction pointer.

        Currently used internally. Must set on a instruction boundary.
        """
        self._ncs.seek(v)

    def _exec(self, i):
        """
        Execute a single instruction.
        """

        # NB: ip has already incremented at this point
        this_ip = self.ip - 2 - i.extra_len

        if i.op == Opcode.NO_OPERATION:
            pass

        elif i.op == Opcode.JMP:
            _logger.debug("JMP %d+%d -> %d", this_ip, i.extra[0], this_ip + i.extra[0])
            self.ip = this_ip + i.extra[0]
            return

        elif i.op == Opcode.JSR:
            if (
                self._stop_before_main
                and this_ip + i.extra[0]
                == self._ndb.function_by_name("main").b_start - 8 - 4 - 1
            ):
                self._stop_before_main = False
                _logger.debug("Stopping before main (UNTESTED)")
                self._stack.pop_int()  # bp
                raise _ScriptExecutionEnd()

            self._ret.append(self.ip)
            self.ip = this_ip + i.extra[0]
            return

        elif i.op == Opcode.JZ:
            if self._stack.pop_int() == 0:
                _logger.debug(
                    "JZ %d+%d -> %d", this_ip, i.extra[0], this_ip + i.extra[0]
                )
                self.ip = this_ip + i.extra[0]
                return

        elif i.op == Opcode.JNZ:
            if self._stack.pop_int() != 0:
                self.ip += i.extra[0]
                return

        elif i.op == Opcode.RET:
            if len(self._ret) == 0:
                raise _ScriptExecutionEnd()
            self.ip = self._ret.pop()
            return

        elif i.op == Opcode.SAVE_BASE_POINTER:
            self._stack.push_int(self._bp)
            self._bp = self.sp

        elif i.op == Opcode.RESTORE_BASE_POINTER:
            self._bp = self._stack.pop_int()

        elif i.op == Opcode.RUNSTACK_ADD:
            if i.aux == Auxcode.TYPE_INTEGER:
                self._stack.push_int(0)
            elif i.aux == Auxcode.TYPE_FLOAT:
                self._stack.push_float(0.0)
            elif i.aux == Auxcode.TYPE_STRING:
                self._stack.push_string("")
            elif i.aux == Auxcode.TYPE_OBJECT:
                self._stack.push_object(Object.INVALID)
            else:
                raise NotImplementedError(f"RUNSTACK_ADD {i.aux}")

        elif i.op in [Opcode.RUNSTACK_COPY, Opcode.RUNSTACK_COPY_BASE]:
            stack_loc = (self._bp if i.op == Opcode.RUNSTACK_COPY_BASE else self.sp) - (
                -1 * i.extra[0]
            ) // 4
            copy_size = i.extra[1] // 4

            for j in range(copy_size):
                self._stack.assign(stack_loc + j, self.sp + j)

        elif i.op in [Opcode.ASSIGNMENT, Opcode.ASSIGNMENT_BASE]:
            stack_loc = (self._bp if i.op == Opcode.ASSIGNMENT_BASE else self.sp) - (
                -1 * i.extra[0]
            ) // 4
            copy_size = i.extra[1] // 4

            for j in range(copy_size):
                self._stack.assign(self.sp - copy_size + j, stack_loc + j)

        elif i.op == Opcode.CONSTANT:
            if i.aux == Auxcode.TYPE_INTEGER:
                self._stack.push_int(i.extra[0])
            elif i.aux == Auxcode.TYPE_FLOAT:
                self._stack.push_float(i.extra[0])
            elif i.aux == Auxcode.TYPE_STRING:
                self._stack.push_string(i.extra[0])
            elif i.aux == Auxcode.TYPE_OBJECT:
                match i.extra[0]:
                    case 0:  # SELF
                        self._stack.push_object(self._object_self)
                    case 1:  # INVALID
                        self._stack.push_object(Object.INVALID)
                    case _:
                        raise NotImplementedError(f"CONSTANT OBJECT {i.extra[0]}")
            else:
                raise NotImplementedError(f"CONSTANT {i.aux}")

        elif i.op == Opcode.MODIFY_STACK_POINTER:
            stack_loc = -1 * i.extra[0] // 4
            self._stack.set_stack_pointer(self.sp - stack_loc)

        elif i.op in [
            Opcode.INCREMENT,
            Opcode.DECREMENT,
            Opcode.INCREMENT_BASE,
            Opcode.DECREMENT_BASE,
        ]:
            stack_loc = (
                self._bp
                if i.op in [Opcode.INCREMENT_BASE, Opcode.DECREMENT_BASE]
                else self.sp
            ) - (-1 * i.extra[0] // 4)
            delta = 1 if i.op in [Opcode.INCREMENT, Opcode.INCREMENT_BASE] else -1

            self._stack.adjust(stack_loc, delta)

        elif i.op == Opcode.NEGATION:
            if i.aux == Auxcode.TYPE_INTEGER:
                self._stack.push_int(-1 * self._stack.pop_int())
            elif i.aux == Auxcode.TYPE_FLOAT:
                self._stack.push_float(-1 * self._stack.pop_float())
            else:
                raise NotImplementedError()

        elif i.op in [
            Opcode.EQUAL,
            Opcode.NOT_EQUAL,
            Opcode.LT,
            Opcode.GT,
            Opcode.LEQ,
            Opcode.GEQ,
        ]:
            # We can probably have the op logic here instead of the stack,
            # then we can turn the stack into a thinner wrapper for perf

            b = self._stack.pop()
            a = self._stack.pop()
            if i.op == Opcode.EQUAL:
                self._stack.push_int_bool(a == b)
            elif i.op == Opcode.NOT_EQUAL:
                self._stack.push_int_bool(a != b)
            elif i.op == Opcode.LT:
                self._stack.push_int_bool(a < b)
            elif i.op == Opcode.GT:
                self._stack.push_int_bool(a > b)
            elif i.op == Opcode.LEQ:
                self._stack.push_int_bool(a <= b)
            elif i.op == Opcode.GEQ:
                self._stack.push_int_bool(a >= b)
            else:
                raise NotImplementedError()

        elif i.op == Opcode.LOGICAL_OR:
            assert i.aux == Auxcode.TYPETYPE_INTEGER_INTEGER
            b = self._stack.pop_int() != 0
            a = self._stack.pop_int() != 0
            self._stack.push_int_bool(a or b)

        elif i.op == Opcode.LOGICAL_AND:
            assert i.aux == Auxcode.TYPETYPE_INTEGER_INTEGER
            b = self._stack.pop_int() != 0
            a = self._stack.pop_int() != 0
            self._stack.push_int_bool(a and b)

        elif i.op == Opcode.INCLUSIVE_OR:
            assert i.aux == Auxcode.TYPETYPE_INTEGER_INTEGER
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a | b)

        elif i.op == Opcode.BOOLEAN_AND:
            assert i.aux == Auxcode.TYPETYPE_INTEGER_INTEGER
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a & b)

        elif i.op == Opcode.EXCLUSIVE_OR:
            assert i.aux == Auxcode.TYPETYPE_INTEGER_INTEGER
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a ^ b)

        elif i.op == Opcode.BOOLEAN_NOT:
            self._stack.push_int_bool(not self._stack.pop_int_bool())

        elif i.op == Opcode.ONES_COMPLEMENT:
            self._stack.push_int(~self._stack.pop_int())

        elif i.op == Opcode.SHIFT_LEFT:
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a << b)

        elif i.op == Opcode.SHIFT_RIGHT:
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a >> b)

        elif i.op == Opcode.USHIFT_RIGHT:
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a >> b)

        elif i.op == Opcode.ADD:
            b = self._stack.pop()
            a = self._stack.pop()
            self._stack.push(a + b)

        elif i.op == Opcode.SUB:
            b = self._stack.pop()
            a = self._stack.pop()
            self._stack.push(a - b)

        elif i.op == Opcode.MUL:
            b = self._stack.pop()
            a = self._stack.pop()
            self._stack.push(a * b)

        elif i.op == Opcode.DIV:
            b = self._stack.pop()
            a = self._stack.pop()
            self._stack.push(a / b)

        elif i.op == Opcode.MODULUS:
            assert i.aux == Auxcode.TYPETYPE_INTEGER_INTEGER
            b = self._stack.pop_int()
            a = self._stack.pop_int()
            self._stack.push_int(a % b)

        elif i.op == Opcode.DE_STRUCT:
            size_orig = i.extra[0] // 4
            start = i.extra[1] // 4
            size = i.extra[2] // 4

            if size + start < size_orig:
                self.sp -= size_orig
                self.sp += size + start
                self._stack.set_stack_pointer(self.sp)

            if start > 0:
                for j in range(self.sp - size - start, self.sp - start):
                    self._stack.assign(j + start, j)

                self.sp -= start
                self._stack.set_stack_pointer(self.sp)

        elif i.op == Opcode.STORE_IP:
            self._save_ip = self.ip + i.aux.value

        elif i.op == Opcode.STORE_STATE:
            self._save_ip = self.ip + i.aux.value
            self._save_bp = i.extra[0]
            self._save_sp = i.extra[1]

        elif i.op == Opcode.EXECUTE_COMMAND:
            cmd = i.extra[0]
            argc = i.extra[1]

            if cmd >= len(self._spec.functions):
                raise ValueError(f"Command {cmd} not implemented")

            vmcmd = self._spec.functions[cmd]

            if not hasattr(self._impl, vmcmd.name):
                raise ValueError(f"Method {vmcmd.name} not implemented")

            if argc != len(vmcmd.args):
                raise ValueError(f"Expected {len(vmcmd.name)} arguments, got {argc}")

            method = getattr(self._impl, vmcmd.name)

            if not callable(method):
                raise TypeError(f"Method {vmcmd.name} is not a callable")

            args = []
            for expect in vmcmd.args:
                if expect.ty != VMType.ACTION:
                    args.append(self._stack.pop_vmtype(expect.ty))
                else:
                    args.append(
                        _VMClosure(ip=self._save_ip, sp=self._save_sp, bp=self._save_bp)
                    )

            rv = method(*args)

            _logger.debug("CMD %s(%s) -> %s", vmcmd.name, args, rv)

            if vmcmd.return_type != VMType.VOID:
                self._stack.push_vmtype(vmcmd.return_type, rv)

        else:
            raise NotImplementedError(f"Opcode {i.op} not implemented")

        assert self.ip >= 0
        assert self.sp >= 0

    def run(self, from_start=True):
        """
        Start running code.

        Args:
            from_start: If True, start from the beginning of the script.
        """

        if len(self._ret) > 0:
            raise ValueError("In the middle of some execution context?")

        if from_start:
            self.ip = 0
            self._stack.set_stack_pointer(0)

        _logger.debug("RUN ip=%d sp=%d", self.ip, self.sp)

        while True:
            i = read_instr(self._ncs)

            _logger.debug(
                "INSTR %7d +%-4d %-20s %-20s (%s)",
                self.ip - (2 + i.extra_len),
                self.sp,
                i.op.name,
                i.aux.name,
                i.extra,
            )

            try:
                self._exec(i)
            except _ScriptExecutionEnd:
                _logger.debug("DONE")
                break

    def proxy(self):
        """
        Return a proxy object that can directly invoke NWScript functions
        from the python side.

        This will only work if NDB information is present.
        """

        parent = self

        class Proxy:
            def __getattr__(self, name):
                def fn(*args):
                    return parent.call(name, *args)

                return fn

        return Proxy()

    def _push_ndb_type(self, ty, value=None):
        match ty:
            case ndb.ScalarType.VOID:
                pass
            case ndb.ScalarType.INT:
                self._stack.push_int(value)
            case ndb.ScalarType.FLOAT:
                self._stack.push_float(value)
            case ndb.ScalarType.STRING:
                self._stack.push_string(value)
            case ndb.ScalarType.OBJECT:
                self._stack.push_object(value)
            case ndb.ScalarType.EFFECT:
                self._stack.push_vmtype(VMType.EFFECT, value)
            case ndb.StructRef(struct_id):
                # Structs are not actually a real datatype in nwscript bytecode:
                # they are pushed/popped as individual stack elements.
                # The python representation is always a dict of field name -> value;
                # except for the builtin "vector" struct, which is a NamedTuple.
                struct = self._ndb.structs[struct_id]
                for lbl, fty in struct.fields:
                    # The code here works for both tuples and dicts.
                    lval = (
                        (value[lbl] if isinstance(value, dict) else getattr(value, lbl))
                        if value is not None
                        else None
                    )
                    self._push_ndb_type(fty, lval)
            case _:
                raise NotImplementedError(f"Type {ty} not implemented")

    def _pop_ndb_type(self, ty, idx=-1):
        match ty:
            case ndb.ScalarType.VOID:
                return None
            case ndb.ScalarType.INT:
                return self._stack.pop_int(idx)
            case ndb.ScalarType.FLOAT:
                return self._stack.pop_float(idx)
            case ndb.ScalarType.STRING:
                return self._stack.pop_string(idx)
            case ndb.ScalarType.OBJECT:
                return self._stack.pop_object(idx)
            case ndb.ScalarType.EFFECT:
                return self._stack.pop_vmtype(VMType.EFFECT)
            case ndb.StructRef(struct_id):
                struct = self._ndb.structs[struct_id]
                values = {
                    lbl: self._pop_ndb_type(fty, idx) for lbl, fty in struct.fields
                }
                # Special case: the builtin "vector" is a NamedTuple as per _types;
                # This is the only one that is not dict.
                if struct.label == "vector":
                    return Vector(**values)
                return values
            case _:
                raise NotImplementedError(f"Type {ty} not implemented")

    def call(self, fn, *args):
        """
        Call a NWScript function.

        This will only work if NDB information is present.

        Args:
            fn: The function name.
            args: The arguments to the function.

        Raises:
            ValueError: If the function is not found, or if the number of arguments
                does not match the function signature.
        """

        assert self._ndb
        if len(self._ret) > 0:
            raise ValueError(f"Ret stack not empty {self._ret}")

        f = self._ndb.function_by_name(fn)

        if f.b_start >= 0xFFFFFFFF:
            raise ValueError(f"Function {fn} was elided out")
        self.ip = f.b_start - 8 - 4 - 1

        _logger.debug("CALL fn=%s", fn)

        self._push_ndb_type(f.ret_type)

        if len(args) != len(f.args):
            raise ValueError(f"Expected {len(f.args)} arguments, got {len(args)}")

        for arg, argtype in zip(args, f.args):
            self._push_ndb_type(argtype, arg)

        self.run(from_start=False)

        return self._pop_ndb_type(f.ret_type, 0)



"""Shared types and helpers useful across the whole library."""

from enum import IntEnum, StrEnum
from dataclasses import dataclass
from typing import NamedTuple


class Language(IntEnum):
    """Maps engine language IDs."""

    ENGLISH = 0
    FRENCH = 1
    GERMAN = 2
    ITALIAN = 3
    SPANISH = 4
    POLISH = 5

    @property
    def code(self) -> str:
        """Get the two-letter language code."""
        return {
            Language.ENGLISH: "en",
            Language.FRENCH: "fr",
            Language.GERMAN: "de",
            Language.ITALIAN: "it",
            Language.SPANISH: "es",
            Language.POLISH: "pl",
        }[self]

    @classmethod
    def from_code(cls, code: str) -> "Language":
        """
        Get the Language enum from a two-letter code.

        Args:
            code: The two-letter language code (eg. "de").

        Returns:
            The corresponding Language enum.

        Raises:
            ValueError: If the code is unknown.
        """
        for lang in cls:
            if lang.code == code.lower():
                return lang
        raise ValueError(f"Unknown language code: {code}")


class CodePage(StrEnum):
    """Maps engine code page names."""

    CP1250 = "cp1250"
    CP1251 = "cp1251"
    CP1252 = "cp1252"

    @classmethod
    def from_language(cls, lang: Language) -> "CodePage":
        """
        Get the default codepage for a given language.

        Args:
            lang: The Language enum.

        Returns:
            The corresponding CodePage enum.

        Raises:
            ValueError: If the language is unknown.
        """
        match lang:
            case (
                Language.ENGLISH
                | Language.FRENCH
                | Language.GERMAN
                | Language.ITALIAN
                | Language.SPANISH
            ):
                return cls.CP1252
            case Language.POLISH:
                return cls.CP1250
        raise ValueError(f"Unknown language: {lang}")


class Gender(IntEnum):
    """Maps engine gender IDs."""

    MALE = 0
    FEMALE = 1


@dataclass(frozen=True)
class GenderedLanguage:
    """
    A combination of Language and Gender.

    This type is needed for various file formats, where the Language and Gender
    types are combined into a single integer.
    """

    lang: Language
    gender: Gender

    def __str__(self):
        return f"{self.lang.name} {self.gender.name}"

    @classmethod
    def from_id(cls, combined_id: int):
        """
        Create a new GenderedLanguage instance from a combined ID.

        Args:
            combined_id: The combined ID, which is 2 times the Language ID plus the Gender.

        Returns:
            A new GenderedLanguage instance.
        """
        lang = Language(combined_id // 2)
        gender = Gender(combined_id % 2)
        return GenderedLanguage(lang, gender)

    def to_id(self) -> int:
        """
        Convert the Language instance to a combined ID.

        Returns:
            The combined ID, which is 2 times the Language ID plus the Gender.
        """
        return self.lang * 2 + self.gender


class FileMagic(bytes):
    """
    A file magic identifies a file type: For NWN, it is the first four characters
    on certain file types.

    Args:
        value: Exactly 4 characters or bytes.

    Raises:
        ValueError: If the input is not exactly 4 bytes long.
    """

    def __new__(cls, value: str | bytes | memoryview):
        if isinstance(value, str):
            value = value.encode("ascii")
        if isinstance(value, memoryview):
            value = value.tobytes()
        if len(value) > 4:
            raise ValueError("Magic must be at most 4 bytes long")
        value = value.ljust(4, b" ")
        if not all(c in b"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 " for c in value):
            raise ValueError(
                "Magic must contain only ASCII uppercase letters, digits, or spaces"
            )
        return super().__new__(cls, value)


class Platform(IntEnum):
    """
    Game platform IDs, as used in various file formats and network messages.
    """

    INVALID = 0
    WINDOWS_X86 = 1
    WINDOWS_X64 = 2

    LINUX_X86 = 10
    LINUX_X64 = 11
    LINUX_ARM32 = 12
    LINUX_ARM64 = 13

    MAC_X86 = 20
    MAC_X64 = 21
    MAC_ARM64 = 22

    IOS = 30

    ANDROID_ARM32 = 40
    ANDROID_ARM64 = 41
    ANDROID_X64 = 42

    NINTENDO_SWITCH = 50

    MICROSOFT_XBOXONE = 60

    SONY_PS4 = 70


class Vector(NamedTuple):
    """
    A 3D vector in the game world.

    Vectors are always immutable, just like in the game engine.
    """

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0



"""
Functionality related to game and user/system environment,
such as locating installation directories and detecting
language and codepage.

Note that most functions here rely on caching to avoid repeated
filesystem access; if you modify environment variables or
configuration files during runtime, you may need to clear
the relevant function caches using functools.cache_clear(). This
will typically only be necessary during testing, is not considered
part of the public API, and will likely change in a future version.
"""

import os
import tomllib
import configparser
import locale
import platform
from pathlib import Path
from typing import Any
from enum import StrEnum

# I'm not overly fond of this. The internal cache system is currently necessary
# for performance, but it might be replaced in the future with more explicit
# control over caching. Caveat empor if you decide to use it in your code.
from functools import cache

from .types import Language, CodePage


class Alias(StrEnum):
    HD0 = "_hd0"  # NWN_HOME
    MODULES = "modules"
    SAVES = "saves"
    OVERRIDE = "override"
    DEVELOPMENT = "development"
    HAK = "hak"
    SCREENSHOTS = "screenshots"
    CURRENTGAME = "currentgame"
    LOGS = "logs"
    TEMP = "temp"
    TEMPCLIENT = "tempclient"
    PATCH = "patch"
    LOCALVAULT = "localvault"
    DMVAULT = "dmvault"
    SERVERVAULT = "servervault"
    OLDSERVERVAULT = "oldservervault"
    DATABASE = "database"
    PORTRAITS = "portraits"
    AMBIENT = "ambient"
    MOVIES = "movies"
    MUSIC = "music"
    TLK = "tlk"
    NWSYNC = "nwsync"
    CACHE = "cache"
    MODELCOMPILER = "modelcompiler"
    CRASHREPORT = "crashreport"

    HD0INSTALL = "_hd0install"  # NWN_ROOT
    PORTRAITSINSTALL = "data/prt"
    MOVIESINSTALL = "data/mov"
    MUSICINSTALL = "data/mus"
    AMBIENTINSTALL = "data/amb"
    PATCHINSTALL = "data/pat"
    PREMIUMINSTALL = "data/prem"
    MODULESINSTALL = "data/mod"
    NWMINSTALL = "data/nwm"
    LCVAULTINSTALL = "data/lcv"
    DMVAULTINSTALL = "data/dmv"
    HAKINSTALL = "data/hk"
    TLKINSTALL = "data/tlk"

    HD0LOCINSTALL = "_hd0locinstall"  # NWN_ROOT / lang
    MOVIESLOCINSTALL = "data/mov"
    TLKLOCINSTALL = "data/tlk"
    OVERRIDELOCINSTALL = "data/ovr"

    @classmethod
    def by_name(cls, name: str) -> "Alias":
        """
        Get the Alias enum from a string name.

        Args:
            name: The alias name (e.g. "MODULES").

        Returns:
            The corresponding Alias enum.

        Raises:
            ValueError: If the name is unknown.
        """
        for alias in cls:
            if alias.value.upper() == name.upper():
                return alias
        raise ValueError(f"{name} is not a valid Alias")


@cache
def _get_settings_tml() -> dict:
    try:
        user = get_user_directory()
        settings_path = os.path.join(user, "settings.tml")
        with open(settings_path, "rb") as settings:
            return tomllib.load(settings)
    except FileNotFoundError:
        pass
    return {}


@cache
def _get_aliases() -> dict:
    try:
        user = get_user_directory()
        ini_path = os.path.join(user, "nwn.ini")
        config = configparser.ConfigParser()
        config.read(ini_path)
        return {k.upper(): v for k, v in config["Alias"].items()}
    except KeyError:
        return {}


@cache
def resolve_alias(alias: str | Alias) -> Path:
    """
    Get the path for a given NWN alias from nwn.ini, or return the
    default value if not found.

    Args:
        alias: The alias to look up.

    Returns:
        The resolved path for the alias, or the default value.

    Raises:
        FileNotFoundError: If the user directory cannot be found. This function
            requires a user directory set up already.
    """

    if isinstance(alias, str):
        alias = Alias.by_name(alias)

    if alias == Alias.HD0:
        return get_user_directory()
    if alias == Alias.HD0INSTALL:
        return get_install_directory()
    if alias == Alias.HD0LOCINSTALL:
        return get_install_language_directory()

    value = alias.value.lower()

    if alias.name.endswith("LOCINSTALL"):
        return get_install_language_directory() / "data" / value

    if alias.name.endswith("INSTALL"):
        return get_install_directory() / "data" / value

    return get_user_directory() / _get_aliases().get(alias.name.upper(), value)


@cache
def get_setting(key: str) -> Any:
    """
    Get a setting from the user's settings.tml file.

    Args:
        key: The setting key to look up, in dot notation (e.g. "game.language.override").

    Returns:
        The setting value, or None if not found.

    Raises:
        FileNotFoundError: If the user directory cannot be found. This function
            requires a user directory set up already.
    """
    settings = _get_settings_tml()
    parts = key.split(".")
    current = settings
    for part in parts:
        if isinstance(current, dict) and part in current:
            current = current[part]
        else:
            return None
    return current


@cache
def get_user_directory() -> Path:
    """
    Find the current user NWN directory.

    The search can be overridden by setting the NWN_HOME or NWN_USER_DIRECTORY
    environment variables.

    Returns:
        The path to the NWN user directory.

    Raises:
        FileNotFoundError: If the directory cannot be found.
    """

    candidates = [
        os.environ.get("NWN_HOME"),
        os.environ.get("NWN_USER_DIRECTORY"),
    ]

    # We never expose system dirs to pytest, just to avoid it clobbering
    # user data and to make tests reproducible.
    if not os.environ.get("PYTEST_CURRENT_TEST"):
        if platform.system() in ("Windows", "Darwin"):
            candidates.append(os.path.expanduser("~/Documents/Neverwinter Nights"))
        else:
            candidates.append(os.path.expanduser("~/.local/share/Neverwinter Nights"))

    for candidate in candidates:
        if candidate and os.path.isdir(candidate):
            return Path(candidate)

    raise FileNotFoundError("Could not locate NWN user directory; try setting NWN_HOME")


@cache
def get_install_directory() -> Path:
    """
    Find the first matching NWN installation directory.
    Currently only supports searching for Steam installs.

    The search can be overridden by setting the NWN_ROOT environment variable.

    Returns:
        The path to the NWN root directory.

    Raises:
        FileNotFoundError: If the directory cannot be found.
    """

    candidates = [
        os.environ.get("NWN_ROOT"),
    ]

    # We never expose system dirs to pytest, just to avoid it clobbering
    # user data and to make tests reproducible.
    if not os.environ.get("PYTEST_CURRENT_TEST"):
        suffix = "/Steam/steamapps/common/Neverwinter Nights/"
        if platform.system() == "Windows":
            candidates.append("c:/program files (x86)/" + suffix)
        elif platform.system() == "Darwin":
            candidates.append(
                os.path.expanduser("~/Library/Application Support/" + suffix)
            )
        else:
            candidates.append(os.path.expanduser("~/.local/share/" + suffix))

    for candidate in candidates:
        if candidate and os.path.isdir(candidate):
            return Path(candidate)

    raise FileNotFoundError("Could not locate NWN; try setting NWN_ROOT")


def get_install_language_directory() -> Path:
    """
    Get the path to the language data directory inside the NWN installation.

    Returns:
        The path to the language data directory.

    Raises:
        FileNotFoundError: If the installation directory cannot be found.
    """
    install = get_install_directory()
    language = get_language()
    return install / "lang" / language.code


@cache
def get_codepage() -> CodePage:
    """
    Detect the codepage used by the NWN installation based on the user
    language.

    The codepage can be overridden by setting the NWN_CODEPAGE environment variable.

    Returns:
        The codepage specified in settings.tml, or defaults based on language.
    """
    try:
        cp = os.environ.get("NWN_CODEPAGE", get_setting("game.language.codepage"))
    except FileNotFoundError:
        cp = None
    try:
        if not cp:
            cp = CodePage.from_language(get_language())
        return CodePage(cp)
    except ValueError:
        return CodePage.CP1252


@cache
def get_language() -> Language:
    """
    Detect the language of the NWN installation based on system locale
    or configuration inside game settings.

    Returns:
        The detected Language enum value, or english by default.
    """

    lang = os.environ.get("NWN_LANGUAGE")

    if not lang:
        try:
            lang = get_setting("game.language.override")
        except FileNotFoundError:
            lang = None

    if not lang:
        if loc := locale.getlocale()[0]:
            lang = loc.split("_")[0]

    try:
        return Language.from_code(lang or "en")
    except ValueError:
        return Language.ENGLISH



"""
Disassemble NCS (compiled script) files.

While the language bytecode is principally game-agnostic, the supported opcodes
describe the default NWN:EE instruction set.
"""

import struct

from enum import Enum
from typing import NamedTuple, BinaryIO
import json

from nwn.environ import get_codepage

_NCS_HEADER = b"NCS V1.0"


class Opcode(Enum):
    ASSIGNMENT = 0x01
    RUNSTACK_ADD = 0x02
    RUNSTACK_COPY = 0x03
    CONSTANT = 0x04
    EXECUTE_COMMAND = 0x05
    LOGICAL_AND = 0x06
    LOGICAL_OR = 0x07
    INCLUSIVE_OR = 0x08
    EXCLUSIVE_OR = 0x09
    BOOLEAN_AND = 0x0A
    EQUAL = 0x0B
    NOT_EQUAL = 0x0C
    GEQ = 0x0D
    GT = 0x0E
    LT = 0x0F
    LEQ = 0x10
    SHIFT_LEFT = 0x11
    SHIFT_RIGHT = 0x12
    USHIFT_RIGHT = 0x13
    ADD = 0x14
    SUB = 0x15
    MUL = 0x16
    DIV = 0x17
    MODULUS = 0x18
    NEGATION = 0x19
    ONES_COMPLEMENT = 0x1A
    MODIFY_STACK_POINTER = 0x1B
    STORE_IP = 0x1C
    JMP = 0x1D
    JSR = 0x1E
    JZ = 0x1F
    RET = 0x20
    DE_STRUCT = 0x21
    BOOLEAN_NOT = 0x22
    DECREMENT = 0x23
    INCREMENT = 0x24
    JNZ = 0x25
    ASSIGNMENT_BASE = 0x26
    RUNSTACK_COPY_BASE = 0x27
    DECREMENT_BASE = 0x28
    INCREMENT_BASE = 0x29
    SAVE_BASE_POINTER = 0x2A
    RESTORE_BASE_POINTER = 0x2B
    STORE_STATE = 0x2C
    NO_OPERATION = 0x2D


class Auxcode(Enum):
    NONE = 0x00
    TYPE_VOID = 0x01
    TYPE_COMMAND = 0x02
    TYPE_INTEGER = 0x03
    TYPE_FLOAT = 0x04
    TYPE_STRING = 0x05
    TYPE_OBJECT = 0x06
    TYPE_ENGST0 = 0x10
    TYPE_ENGST1 = 0x11
    TYPE_ENGST2 = 0x12
    TYPE_ENGST3 = 0x13
    TYPE_ENGST4 = 0x14
    TYPE_ENGST5 = 0x15
    TYPE_ENGST6 = 0x16
    TYPE_ENGST7 = 0x17
    TYPE_ENGST8 = 0x18
    TYPE_ENGST9 = 0x19
    TYPETYPE_INTEGER_INTEGER = 0x20
    TYPETYPE_FLOAT_FLOAT = 0x21
    TYPETYPE_OBJECT_OBJECT = 0x22
    TYPETYPE_STRING_STRING = 0x23
    TYPETYPE_STRUCT_STRUCT = 0x24
    TYPETYPE_INTEGER_FLOAT = 0x25
    TYPETYPE_FLOAT_INTEGER = 0x26
    TYPETYPE_ENGST0_ENGST0 = 0x30
    TYPETYPE_ENGST1_ENGST1 = 0x31
    TYPETYPE_ENGST2_ENGST2 = 0x32
    TYPETYPE_ENGST3_ENGST3 = 0x33
    TYPETYPE_ENGST4_ENGST4 = 0x34
    TYPETYPE_ENGST5_ENGST5 = 0x35
    TYPETYPE_ENGST6_ENGST6 = 0x36
    TYPETYPE_ENGST7_ENGST7 = 0x37
    TYPETYPE_ENGST8_ENGST8 = 0x38
    TYPETYPE_ENGST9_ENGST9 = 0x39
    TYPETYPE_VECTOR_VECTOR = 0x3A
    TYPETYPE_VECTOR_FLOAT = 0x3B
    TYPETYPE_FLOAT_VECTOR = 0x3C
    EVAL_INPLACE = 0x70
    EVAL_POSTPLACE = 0x71


class Instr(NamedTuple):
    op: Opcode
    aux: Auxcode
    extra: tuple
    extra_len: int

    def __str__(self):
        return f"{self.op.name}.{self.aux.name}{self.extra}"

    def shortcode(self) -> str:
        """
        Returns a short code for the instruction, suitable for use in
        disassembly listings.
        """
        aux_code = _AUXCODE_TO_CANONICAL.get(self.aux, "")
        op_code = _OPCODE_TO_CANONICAL.get(self.op, "")
        return f"{op_code}{('.' + aux_code) if aux_code else ''}"


def read_extra(file: BinaryIO, op: Opcode, aux: Auxcode) -> tuple:
    """
    Parses additional data based on the opcode and auxiliary code.

    Note: This will re-parse embedded extra data every time it's called:
    For example, JSON payloads will be evaluated.

    Args:
        file: A binary stream from which to read the data.
        op: The opcode of the instruction.
        aux: The auxiliary code providing additional context for the operation.

    Returns:
        A variable length tuple containing the parsed data, or empty.

    Raises:
        struct.error: If the data cannot be read from the stream.
        ValueError: Various error conditions.
    """

    if op == Opcode.CONSTANT:
        if aux == Auxcode.TYPE_INTEGER:
            return struct.unpack(">i", file.read(4))
        if aux == Auxcode.TYPE_FLOAT:
            return struct.unpack(">f", file.read(4))
        if aux == Auxcode.TYPE_OBJECT:
            return struct.unpack(">i", file.read(4))
        if aux == Auxcode.TYPE_STRING:
            slen = struct.unpack(">H", file.read(2))[0]
            return (file.read(slen).decode(get_codepage()),)
        if aux == Auxcode.TYPE_ENGST2:
            return struct.unpack(">i", file.read(4))
        if aux == Auxcode.TYPE_ENGST7:
            # TODO: this is wasteful, doing it every time we parse an instr
            slen = struct.unpack(">H", file.read(2))[0]
            js = file.read(slen)
            return (json.loads(js),)
        return ()
    if op in [Opcode.JZ, Opcode.JMP, Opcode.JSR, Opcode.JNZ]:
        return struct.unpack(">i", file.read(4))
    if op == Opcode.STORE_STATE:
        return struct.unpack(">ii", file.read(8))
    if op == Opcode.MODIFY_STACK_POINTER:
        return struct.unpack(">i", file.read(4))
    if op == Opcode.EXECUTE_COMMAND:
        return struct.unpack(">HB", file.read(3))
    if op in [Opcode.RUNSTACK_COPY, Opcode.RUNSTACK_COPY_BASE]:
        return struct.unpack(">iH", file.read(6))
    if op in [Opcode.ASSIGNMENT, Opcode.ASSIGNMENT_BASE]:
        return struct.unpack(">iH", file.read(6))
    if op in [
        Opcode.INCREMENT,
        Opcode.DECREMENT,
        Opcode.INCREMENT_BASE,
        Opcode.DECREMENT_BASE,
    ]:
        return struct.unpack(">i", file.read(4))
    if op == Opcode.DE_STRUCT:
        return struct.unpack(">HHH", file.read(6))
    return ()


def read_instr(file: BinaryIO) -> Instr:
    """
    Reads an instruction from the given input stream.

    Args:
        file: A binary stream from which the instruction is read.

    Returns:
        The instruction read from the input stream.

    Raises:
        struct.error: If the data cannot be read from the stream.
        ValueError: If the opcode or auxiliary code is invalid.
    """

    op, aux = struct.unpack(">BB", file.read(2))
    op, aux = Opcode(op), Auxcode(aux)
    io_start = file.tell()
    ex = read_extra(file, op, aux)
    exl = file.tell() - io_start
    return Instr(op, aux, ex, exl)


def disasm(file: BinaryIO):
    """
    Disassembles bytecode from a file-ish.

    Skips any encountered NCS header as a convenience.

    Args:
        file: A file-ish containing the bytecode to disassemble.

    Yields:
        Instr: The next instruction read from the bytecode stream.
    """

    # Skip header as a convenience
    if file.peek(8)[0:8] == _NCS_HEADER:
        file.seek(file.tell() + 8 + 1 + 4)

    while file.peek(1):
        yield read_instr(file)


_OPCODE_TO_CANONICAL = {
    Opcode.ASSIGNMENT: "CPDOWNSP",
    Opcode.RUNSTACK_ADD: "RSADD",
    Opcode.RUNSTACK_COPY: "CPTOPSP",
    Opcode.CONSTANT: "CONST",
    Opcode.EXECUTE_COMMAND: "ACTION",
    Opcode.LOGICAL_AND: "LOGAND",
    Opcode.LOGICAL_OR: "LOGOR",
    Opcode.INCLUSIVE_OR: "INCOR",
    Opcode.EXCLUSIVE_OR: "EXCOR",
    Opcode.BOOLEAN_AND: "BOOLAND",
    Opcode.EQUAL: "EQUAL",
    Opcode.NOT_EQUAL: "NEQUAL",
    Opcode.GEQ: "GEQ",
    Opcode.GT: "GT",
    Opcode.LT: "LT",
    Opcode.LEQ: "LEQ",
    Opcode.SHIFT_LEFT: "SHLEFT",
    Opcode.SHIFT_RIGHT: "SHRIGHT",
    Opcode.USHIFT_RIGHT: "USHRIGHT",
    Opcode.ADD: "ADD",
    Opcode.SUB: "SUB",
    Opcode.MUL: "MUL",
    Opcode.DIV: "DIV",
    Opcode.MODULUS: "MOD",
    Opcode.NEGATION: "NEG",
    Opcode.ONES_COMPLEMENT: "COMP",
    Opcode.MODIFY_STACK_POINTER: "MOVSP",
    Opcode.STORE_IP: "STOREIP",
    Opcode.JMP: "JMP",
    Opcode.JSR: "JSR",
    Opcode.JZ: "JZ",
    Opcode.RET: "RET",
    Opcode.DE_STRUCT: "DESTRUCT",
    Opcode.BOOLEAN_NOT: "NOT",
    Opcode.DECREMENT: "DECSP",
    Opcode.INCREMENT: "INCSP",
    Opcode.JNZ: "JNZ",
    Opcode.ASSIGNMENT_BASE: "CPDOWNBP",
    Opcode.RUNSTACK_COPY_BASE: "CPTOPBP",
    Opcode.DECREMENT_BASE: "DECBP",
    Opcode.INCREMENT_BASE: "INCBP",
    Opcode.SAVE_BASE_POINTER: "SAVEBP",
    Opcode.RESTORE_BASE_POINTER: "RESTOREBP",
    Opcode.STORE_STATE: "STORESTATE",
    Opcode.NO_OPERATION: "NOP",
}

_AUXCODE_TO_CANONICAL = {
    Auxcode.TYPE_INTEGER: "I",
    Auxcode.TYPE_FLOAT: "F",
    Auxcode.TYPE_STRING: "S",
    Auxcode.TYPE_OBJECT: "O",
    Auxcode.TYPE_ENGST0: "E0",
    Auxcode.TYPE_ENGST1: "E1",
    Auxcode.TYPE_ENGST2: "E2",
    Auxcode.TYPE_ENGST3: "E3",
    Auxcode.TYPE_ENGST4: "E4",
    Auxcode.TYPE_ENGST5: "E5",
    Auxcode.TYPE_ENGST6: "E6",
    Auxcode.TYPE_ENGST7: "E7",
    Auxcode.TYPE_ENGST8: "E8",
    Auxcode.TYPE_ENGST9: "E9",
    Auxcode.TYPETYPE_INTEGER_INTEGER: "II",
    Auxcode.TYPETYPE_FLOAT_FLOAT: "FF",
    Auxcode.TYPETYPE_OBJECT_OBJECT: "OO",
    Auxcode.TYPETYPE_STRING_STRING: "SS",
    Auxcode.TYPETYPE_STRUCT_STRUCT: "TT",
    Auxcode.TYPETYPE_INTEGER_FLOAT: "IF",
    Auxcode.TYPETYPE_FLOAT_INTEGER: "FI",
    Auxcode.TYPETYPE_ENGST0_ENGST0: "E0E0",
    Auxcode.TYPETYPE_ENGST1_ENGST1: "E1E1",
    Auxcode.TYPETYPE_ENGST2_ENGST2: "E2E2",
    Auxcode.TYPETYPE_ENGST3_ENGST3: "E3E3",
    Auxcode.TYPETYPE_ENGST4_ENGST4: "E4E4",
    Auxcode.TYPETYPE_ENGST5_ENGST5: "E5E5",
    Auxcode.TYPETYPE_ENGST6_ENGST6: "E6E6",
    Auxcode.TYPETYPE_ENGST7_ENGST7: "E7E7",
    Auxcode.TYPETYPE_ENGST8_ENGST8: "E8E8",
    Auxcode.TYPETYPE_ENGST9_ENGST9: "E9E9",
    Auxcode.TYPETYPE_VECTOR_VECTOR: "VV",
    Auxcode.TYPETYPE_VECTOR_FLOAT: "VF",
    Auxcode.TYPETYPE_FLOAT_VECTOR: "FV",
}



"""Read and write ERF (Encapsulated Resource Format) archives."""

import struct
from pathlib import Path
from typing import NamedTuple, BinaryIO, Mapping
from enum import Enum
from datetime import date, timedelta

from .types import FileMagic, GenderedLanguage
from .environ import get_codepage
from .res import restype_to_extension, extension_to_restype


class Reader(Mapping[str, bytes]):
    """
    Class to read and access an ERF archive (MOD, HAK, ERF, ...)

    Example:
        >>> with open("Prelude.mod", "rb") as file:
        ...    erf = Reader(file)
        ...    print(erf.filenames)
        ...    gff_data = erf.read_file("item.uti")
        ...    ...

    Args:
        file: The file object to read from, either a filename or a BinaryIO.
    """

    class Version(Enum):
        V_1_0 = "V1.0"
        # E_1_0 = "E1.0"

    class _Header(NamedTuple):
        file_type: FileMagic
        file_version: "Reader.Version"
        locstr_count: int
        locstr_sz: int
        entry_count: int
        offset_to_locstr: int
        offset_to_keylist: int
        offset_to_reslist: int
        build_year: int
        build_day: int
        description_strref: int

    class Entry(NamedTuple):
        resref: str
        restype: int
        offset: int
        disk_size: int
        uncompressed_size: int

        @property
        def filename(self):
            return f"{self.resref.lower()}.{restype_to_extension(self.restype)}"

    def _seek(self, relative_to_start):
        self._file.seek(self._root_offset + relative_to_start)

    def __init__(self, file: BinaryIO | str | Path, max_entries=65535, max_locstr=100):
        if isinstance(file, (str, Path)):
            self._owns_file = True
            self._file = open(file, "rb")  # pylint: disable=consider-using-with
        else:
            self._owns_file = False
            self._file = file
        self._root_offset = self._file.tell()

        ft = self._file.read(4)
        fv = self.Version(self._file.read(4).decode("ASCII"))
        va = struct.unpack("IIIIIIIII", self._file.read(36))
        self._header = self._Header(FileMagic(ft), fv, *va)

        if self._header.entry_count > max_entries:
            raise ValueError("Too many resources")

        if self._header.locstr_count > max_locstr:
            raise ValueError("Too many localized strings")

        self._seek(self._header.offset_to_locstr)
        loc_str = {
            GenderedLanguage.from_id(lid): st.decode(get_codepage())
            for lid, sz, st in (
                (lid, sz, self._file.read(sz))
                for lid, sz in struct.iter_unpack(
                    "II", self._file.read(self._header.locstr_count * 8)
                )
            )
        }

        self._seek(self._header.offset_to_reslist)
        resources = [
            (offset, disk_size, disk_size)
            for offset, disk_size in struct.iter_unpack(
                "II", self._file.read(self._header.entry_count * 8)
            )
        ]
        self._seek(self._header.offset_to_keylist)
        keys = [
            (
                resref_bytes.rstrip(b"\x00").decode("ASCII"),
                res_type,
            )
            for resref_bytes, _, res_type, _ in struct.iter_unpack(
                "16sIH2s", self._file.read(self._header.entry_count * 24)
            )
        ]

        self._localized_strings = loc_str

        self._files = {
            f"{resref.lower()}.{restype_to_extension(restype)}": self.Entry(
                resref, restype, o, d, u
            )
            for (resref, restype), (o, d, u) in zip(keys, resources)
        }

    def __del__(self):
        if self._owns_file:
            self._file.close()

    @property
    def file_type(self) -> FileMagic:
        """The file type of the ERF archive."""
        return self._header.file_type

    @property
    def build_date(self) -> date:
        """The build date of the ERF archive."""
        return date(1900 + self._header.build_year, 1, 1) + timedelta(
            days=self._header.build_day
        )

    @property
    def localized_strings(self) -> dict[GenderedLanguage, str]:
        """The localized strings in the ERF archive."""
        return self._localized_strings

    @property
    def description_strref(self) -> int:
        """The STRREF set in the ERF header. 0 if not set."""
        return self._header.description_strref

    @property
    def filenames(self) -> list[str]:
        """
        Returns the filenames in the ERF archive.

        Returns:
            A list of filenames present in the ERF archive.
        """
        return list(self._files.keys())

    @property
    def filemap(self) -> dict[str, Entry]:
        """
        Returns the mapping of files.

        This method returns the internal dictionary that maps file names to their
        corresponding file data.

        Returns:
            A dict mapping file names to their corresponding Entry objects.
        """
        return self._files

    def read_file(self, filename: str) -> bytes:
        """
        Retrieve the contents of a file from the archive.

        Args:
            filename: The name of the file to retrieve.

        Returns:
            The contents of the file as a byte string.

        Raises:
            KeyError: If the file is not found in the archive.
            ValueError: If the filename is of a unknown restype.
        """
        resource = self._files[filename.lower()]
        self._seek(resource.offset)
        return self._file.read(resource.disk_size)

    def __getitem__(self, item: str) -> bytes:
        return self.read_file(item)

    def __iter__(self):
        return iter(self.filenames)

    def __len__(self) -> int:
        return len(self._files)


class Writer:
    """
    A class to write ERF files.

    Example:
        >>> with open("Prelude.mod", "wb") as file:
        ...    with Writer(file, file_type="MOD ") as e:
        ...        e.add_localized_string(Language.ENGLISH, "Prelude")
        ...        with open("item.uti", "rb") as item:
        ...            e.add_file("item.uti", item)
    """

    class Entry(NamedTuple):
        resref: str
        restype: int
        offset: int
        size: int

    def __init__(
        self,
        file: BinaryIO,
        file_type: FileMagic = FileMagic(b"ERF "),
        build_date=date.today(),
    ):
        self._file = file
        self._entries = []
        self._locstr = {}

        self._file_type = FileMagic(file_type)
        self._build_year = build_date.year - 1900
        self._build_day = build_date.timetuple().tm_yday - 1

    def __enter__(self):
        self._file.write(self._file_type)
        self._file.write(Reader.Version.V_1_0.value.encode("ASCII"))
        self._file.write(b"\x00" * 36)
        self._file.write(b"\x00" * 116)  # reserved bytes as per spec
        assert self._file.tell() == 160
        return self

    def add_localized_string(self, gendered_lang: GenderedLanguage, text):
        self._locstr[gendered_lang] = text

    def add_file_data(self, filename: str, data: bytes):
        """
        Adds a file to the ERF archive.

        Args:
            filename: The name of the file to add, including its extension.
            data: The binary data of the file to add.
        """
        offset = self._file.tell()
        size = len(data)
        # ensure we have a restype
        base, ext = filename.split(".")
        if len(base) > 16:
            raise ValueError("Resource name too long")
        rt = extension_to_restype(ext)
        self._entries.append(Writer.Entry(base, rt, offset, size))
        self._file.write(data)
        assert self._file.tell() == offset + size

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return False

        locstr_offset = self._file.tell()
        for gendered_lang, text in self._locstr.items():
            encoded_text = text.encode(get_codepage())
            self._file.write(struct.pack("I", gendered_lang.to_id()))
            self._file.write(struct.pack("I", len(encoded_text)))
            self._file.write(encoded_text)
        locstr_size = self._file.tell() - locstr_offset

        keylist_offset = self._file.tell()
        for resref, restype, offset, size in self._entries:
            res_ref = resref.ljust(16, "\x00").encode("ASCII")
            res_id = 0  # res_id unused
            self._file.write(res_ref)
            self._file.write(struct.pack("I", res_id))
            self._file.write(struct.pack("H", restype))
            self._file.write(b"\x00\x00")  # Unused

        reslist_offset = self._file.tell()
        for _, _, offset, size in self._entries:
            self._file.write(struct.pack("i", offset))
            self._file.write(struct.pack("i", size))

        eof_offset = self._file.tell()

        self._file.seek(8)
        self._file.write(
            struct.pack(
                "IIIIIIIII",
                len(self._locstr),
                locstr_size,
                len(self._entries),
                locstr_offset,
                keylist_offset,
                reslist_offset,
                self._build_year,
                self._build_day,
                0,  # description_strref
            )
        )
        self._file.seek(eof_offset)
        return False



from dataclasses import dataclass, field
from typing import Any

from nwn.nwscript import Object, VMType


@dataclass(order=True)
class _Elem:
    # The only reason this is here is basically because we want to
    # support None as a construction value
    type: VMType = field(compare=False)
    value: Any

    def __post_init__(self):
        expect_type = self.type.python_type()

        if expect_type is None:
            # As a convenience, we allow returning None from the VM
            # We must never GIVE the VM a None, though
            raise TypeError(f"Cannot stack type {self.type}")

        if self.value is None:
            self.value = expect_type()

        elif not isinstance(self.value, expect_type):
            raise TypeError(f"Value {self.value} is not of type {expect_type}")

        if isinstance(self.value, int):
            self.value = (self.value + 2**31) % (2**32) - 2**31

    def __str__(self):
        return f"{self.type.value} {repr(self.value)}"

    def adjust(self, delta):
        if isinstance(self.value, int) and isinstance(delta, int):
            self.value += delta
            self.value = (self.value + 2**31) % (2**32) - 2**31
        elif isinstance(self.value, float) and isinstance(delta, float):
            self.value += delta
        else:
            raise NotImplementedError()

    # Any Int/Float pair auto-coerces to float
    def _op(self, other, op):
        if any(t == VMType.FLOAT for t in [self.type, other.type]):
            return _Elem(VMType.FLOAT, op(float(self.value), float(other.value)))
        if self.type != other.type:
            raise TypeError(f"Op invalid on {self.type} and {other.type}")
        return _Elem(self.type, op(self.value, other.value))

    def __add__(self, other):
        return self._op(other, lambda a, b: a + b)

    def __sub__(self, other):
        return self._op(other, lambda a, b: a - b)

    def __mul__(self, other):
        return self._op(other, lambda a, b: a * b)

    def __truediv__(self, other):
        def mydiv(a, b):
            if isinstance(a, int) and isinstance(b, int):
                return a // b
            else:
                return a / b

        return self._op(other, mydiv)


class Stack:
    """
    A stack as used by the NWN VM.

    You usually do not need to use this directly, but rather use the
    VM class, which uses this internally.
    """

    def __init__(self):
        self._stack = []
        self._sp = 0

    @property
    def sp(self):
        return self._sp

    @sp.setter
    def sp(self, value):
        assert value >= 0
        assert value <= len(self._stack)
        self._sp = value

    # TODO: this should go away (use self.sp property, but needs careful vetting)
    def set_stack_pointer(self, p: int):
        self._sp = p
        self._stack = self._stack[: self.sp]

    def adjust(self, idx, delta):
        self._stack[idx].adjust(delta)

    def assign(self, src: int, dst: int):
        assert src >= 0, f"src: {src}"
        assert dst >= 0, f"dst: {dst}"
        if dst >= len(self._stack):
            self._stack.append(self._stack[src])
            self._sp += 1
        else:
            self._stack[dst] = self._stack[src]

    def push(self, e: _Elem):
        assert isinstance(e, _Elem)
        self._stack.append(e)
        self._sp += 1

    def pop(self, idx=-1) -> _Elem:
        assert self._sp > 0
        self._sp -= 1
        r = self._stack.pop(idx)
        return r

    def push_vmtype(self, ty: VMType, value: Any):
        assert ty != VMType.VECTOR  # TODO
        assert ty != VMType.ACTION
        self.push(_Elem(ty, value))

    def pop_vmtype(self, ty: VMType, idx=-1) -> Any:
        assert ty != VMType.VECTOR  # TODO
        assert ty != VMType.ACTION
        test = self.pop(idx)
        if test.type != ty:
            raise TypeError(f"Expected {ty}, got {test.type}")
        return test.value

    def push_int(self, v: int):
        self.push_vmtype(VMType.INT, v)

    def push_int_bool(self, v: bool):
        self.push_int(1 if v else 0)

    def push_float(self, v: float):
        self.push_vmtype(VMType.FLOAT, v)

    def push_string(self, v: str):
        self.push_vmtype(VMType.STRING, v)

    def push_object(self, v: Object):
        self.push_vmtype(VMType.OBJECT, v)

    def pop_int(self, idx=-1) -> int:
        return self.pop_vmtype(VMType.INT, idx)

    def pop_int_bool(self, idx=-1) -> bool:
        return bool(self.pop_int(idx))

    def pop_float(self, idx=-1) -> float:
        return self.pop_vmtype(VMType.FLOAT, idx)

    def pop_string(self, idx=-1) -> str:
        return self.pop_vmtype(VMType.STRING, idx)

    def pop_object(self, idx=-1) -> Object:
        return self.pop_vmtype(VMType.OBJECT, idx)



"""
Types and classes for ResMan (resource management) functionality.
"""

from typing import Mapping
from collections import UserDict
from abc import ABC


RESTYPE_MAP = {
    0: "res",
    1: "bmp",
    2: "mve",
    3: "tga",
    4: "wav",
    5: "wfx",
    6: "plt",
    7: "ini",
    8: "bmu",
    9: "mpg",
    10: "txt",
    2000: "plh",
    2001: "tex",
    2002: "mdl",
    2003: "thg",
    2005: "fnt",
    2007: "lua",
    2008: "slt",
    2009: "nss",
    2010: "ncs",
    2011: "mod",
    2012: "are",
    2013: "set",
    2014: "ifo",
    2015: "bic",
    2016: "wok",
    2017: "2da",
    2018: "tlk",
    2022: "txi",
    2023: "git",
    2024: "bti",
    2025: "uti",
    2026: "btc",
    2027: "utc",
    2029: "dlg",
    2030: "itp",
    2031: "btt",
    2032: "utt",
    2033: "dds",
    2034: "bts",
    2035: "uts",
    2036: "ltr",
    2037: "gff",
    2038: "fac",
    2039: "bte",
    2040: "ute",
    2041: "btd",
    2042: "utd",
    2043: "btp",
    2044: "utp",
    2045: "dft",
    2046: "gic",
    2047: "gui",
    2048: "css",
    2049: "ccs",
    2050: "btm",
    2051: "utm",
    2052: "dwk",
    2053: "pwk",
    2054: "btg",
    2055: "utg",
    2056: "jrl",
    2057: "sav",
    2058: "utw",
    2059: "4pc",
    2060: "ssf",
    2061: "hak",
    2062: "nwm",
    2063: "bik",
    2064: "ndb",
    2065: "ptm",
    2066: "ptt",
    2067: "bak",
    2068: "dat",
    2069: "shd",
    2070: "xbc",
    2071: "wbm",
    2072: "mtr",
    2073: "ktx",
    2074: "ttf",
    2075: "sql",
    2076: "tml",
    2077: "sq3",
    2078: "lod",
    2079: "gif",
    2080: "png",
    2081: "jpg",
    2082: "caf",
    2083: "jui",
    9996: "ids",
    9997: "erf",
    9998: "bif",
    9999: "key",
    0xFFFF: "___",
}


def restype_to_extension(restype: int) -> str:
    """
    Convert a resource type to its corresponding file extension.

    Args:
        restype: The resource type to convert.

    Returns:
        The corresponding file extension for the given resource type.

    Raises:
        ValueError: If the given resource type is unknown.
    """
    try:
        return RESTYPE_MAP[restype]
    except KeyError as e:
        raise ValueError(f"Unknown restype: {restype}") from e


def extension_to_restype(extension: str) -> int:
    """
    Convert a file extension to its corresponding resource type identifier.

    Args:
        extension: The file extension to convert.

    Returns:
        The resource type identifier corresponding to the given extension.

    Raises:
        ValueError: If the extension is not recognized.
    """

    try:
        return {v: k for k, v in RESTYPE_MAP.items()}[extension.lower()]
    except KeyError as e:
        raise ValueError(f"Unknown extension: {extension}") from e


def is_valid_resref(f: str) -> bool:
    """
    Check if a given filename is a valid NWN resref.

    Args:
        f: The filename to check.

    Returns:
        True if the filename is a valid resref, False otherwise.
    """
    if "/" in f or "\\" in f or f.count(".") != 1:
        return False
    name, ext = f.rsplit(".", 1)
    if len(name) > 16 or not name:
        return False
    try:
        extension_to_restype(ext)
        return True
    except ValueError:
        return False


class Container(Mapping[str, bytes], ABC):
    """
    A generic resource container interface suitable for plugging into ResMan.
    """


class ResDict(Container, UserDict[str, bytes]):
    """
    A case-insensitive resource dict that validates resource references.
    Data is stored in memory.
    """

    def _normalize(self, key: str) -> str:
        return key.lower()

    def __getitem__(self, key: str) -> bytes:
        return self.data[self._normalize(key)]

    def __setitem__(self, key: str, value: bytes):
        if not is_valid_resref(key):
            raise ValueError(f"Invalid resref: {key}")
        self.data[self._normalize(key)] = value

    def __delitem__(self, key: str):
        del self.data[self._normalize(key)]



"""
nwscript-related utilities.
"""

from ..types import Vector
from ._types import (
    Object,
    Location,
    Effect,
    EngineStructure,
    VMType,
)

__all__ = [
    "Object",
    "Vector",
    "Location",
    "Effect",
    "EngineStructure",
    "VMType",
]



"""
A NWN-like resman implementation.

Allows stacking resources origins with ChainMap semantics with similar lookup
behaviour to the NWN engine.

The provided default implementation will correctly look up user-configured
aliases. If you do not want this behaviour, you can create your own
ResMan/ChainMap by combining your own resource containers; or you can override
the env var NWN_ROOT/NWN_HOME to point to a custom installation/user directory.

Example:

    >>> from nwn.erf import Reader as ErfReader
    ... from nwn.res import ResDict
    ... from nwn.resdir import ResDir
    ... from nwn.resman import create as create_resman
    ...
    ... module_erf = ErfReader("mymodule.erf")
    ... inmem = ResDict()
    ... rm = create_resman(inmem, ResDir("local_dir"), module_erf)
    ...
    ... ifo = rm["module.ifo"]         # from mymodule.erf
    ... nss = rm["nwscript.nss"]       # from retail keyfile
    ... rm["temp.txt"] = b"SomeData"   # writes to inmem dict

"""

from collections import ChainMap

from nwn.environ import (
    get_install_directory,
    get_install_language_directory,
    get_user_directory,
    resolve_alias,
    Alias,
)
from nwn.res import Container
from nwn.resdir import LocalDirectory
from nwn.key import Reader as Key


class ResMan(ChainMap[str, bytes]):
    """
    A resource manager that chains multiple resource mappings together.

    This is currently a type alias for ChainMap[str, bytes] to indicate its
    intended use, but may be extended in the future; so make sure to use
    this type for foward compatibility.
    """


def create(*maps: Container, include_user: bool = True) -> ResMan:
    """
    Create a default resman instance containing all base and user game data,
    in addition to any additional resource maps provided. Uses the environment
    to locate NWN installation and user directories.

    Hint: If you want to create a resman that can also store data, prefix
    your maps chain with a ResDict() instance (for in-memory storage or
    a ResDir(..., writable=True) instance pointing to a writable directory).

    Args:
        maps: Additional resource maps to include in the resman, in order of
            precedence (first has highest precedence).
        include_user: Whether to include the user directory aliases.

    Returns:
        A ResMan containing all provided resource maps, plus the base game
        resource keys.

    Raises:
        FileNotFoundError: If any of the required directories cannot be found.
    """
    try:
        user = get_user_directory() if include_user else None
    except FileNotFoundError:
        user = None

    dataroot = get_install_directory() / "data"
    langdataroot = get_install_language_directory() / "data"

    retailkey = dataroot / "nwn_retail.key"
    # Not all installs have a retail.key (eg the headless server package)
    if not retailkey.is_file():
        retailkey = None

    langkey = langdataroot / "nwn_base_loc.key"
    # Not all language overrides have a keyfile
    if not langkey.is_file():
        langkey = None

    def loc_dir(alias: Alias) -> LocalDirectory | None:
        if not user:
            return None
        return LocalDirectory(resolve_alias(alias))

    stack = [
        *maps,
        # TEMPCLIENT: not used for our python impl
        loc_dir(Alias.PORTRAITS),
        loc_dir(Alias.PORTRAITSINSTALL),
        # DMVAULT: not enabled by default
        # LOCALVAULT: not enabled by default
        # SERVERVAULT: not enabled by default
        # DMVAULTINSTALL: not enabled by default
        # LCVAULTINSTALL: not enabled by default
        loc_dir(Alias.DEVELOPMENT),
        loc_dir(Alias.OVERRIDE),
        loc_dir(Alias.OVERRIDELOCINSTALL),
        # OVERRIDEINSTALL: not used since patch 37
        loc_dir(Alias.AMBIENT),
        loc_dir(Alias.AMBIENTINSTALL),
        loc_dir(Alias.MUSIC),
        loc_dir(Alias.MUSICINSTALL),
        Key(langkey) if langkey else None,
        Key(dataroot / "nwn_retail.key") if retailkey else None,
        Key(dataroot / "nwn_base.key"),
    ]
    return ResMan(*[s for s in stack if s is not None])



"""
A VM that can run NWScript bytecode with bidirectional Python linkage.

This module is very, very preliminary. Expect the types to move and change.

Example:
    >>> from nwn.nwscript.vm import VM, Script
    ... from nwn.nwscript import langspec
    ...
    ... class Impl:
    ...     def Random(self, nMax) -> int:
    ...         return nMax + 1
    ...
    ... script = Script.from_compiled("test")  # requires test.{ncs,nss}
    ... with open("nwscript.nss", "r") as nws:
    ...     spec = langspec.read(nws)
    ... vm = VM(script=script, spec=spec, impl=Impl())
    ... proxy = vm.proxy()
    ...
    ... assert vm.some_func(5) == 42
    ... assert vm.some_func2() == {"a": 42, "b": 3.14, "c": "Hellorld"}

Contents of ``test.nss``; compile with ndb (-g):

    .. code-block:: c

        struct Testing { int a; float b; string c; };

        int some_func() { return Random(41); }

        struct Testing some_func2() {
            struct Testing t;
            t.a = 42;
            t.b = 3.14;
            t.c = "Hellorld";
            return t;
        }

        void main() {}

"""

from ._stack import Stack
from ._script import Script
from ._vm import VM

__all__ = [
    "Stack",
    "Script",
    "VM",
]



"""
Read and write soundset files.

Note that the index in the SSF always has a special meaning (soundset.2da),
and cannot be reordered.
"""

import struct
from typing import BinaryIO, NamedTuple


class Entry(NamedTuple):
    """
    A single entry in a SSF file including the resref and strref.
    """

    resref: str = ""
    strref: int = 0xFFFFFFFF


def read(file: BinaryIO, max_entries=0xFF) -> list[Entry]:
    """
    Read a SSF file and return a list of entries.

    Args:
        file: A binary file object containing the SSF file.
        max_entries: The maximum number of entries to read.

    Returns:
        A list of tuples containing the resref and strref of each entry.

    Raises:
        ValueError: If the file does not contain valid SSF data.
    """

    magic = file.read(8)
    if magic != b"SSF V1.0":
        raise ValueError("Not a valid SSF file")

    entry_count = struct.unpack("<I", file.read(4))[0]
    if entry_count > max_entries:
        raise ValueError(f"Too many entries in SSF file: {entry_count} > {max_entries}")

    table_offset = struct.unpack("<I", file.read(4))[0]
    if table_offset != 40:
        raise ValueError("Invalid table offset")

    padding = file.read(24)
    if padding != b"\x00" * 24:
        raise ValueError("Invalid padding")

    entry_offsets = [struct.unpack("<I", file.read(4))[0] for _ in range(entry_count)]
    entries = []
    for offset in entry_offsets:
        file.seek(offset)
        resref = file.read(16).decode("ascii").strip("\x00")
        strref = struct.unpack("<I", file.read(4))[0]
        entries.append(Entry(resref, strref))

    return entries


def write(file: BinaryIO, entries: list[Entry]):
    """
    Writes a list of entries to a binary file in the SSF V1.0 format.

    Args:
        file: The binary file to write to.
        entries: A list of Entry objects to be written to the file.
    """

    file.write(b"SSF V1.0")
    file.write(struct.pack("<I", len(entries)))
    file.write(struct.pack("<I", 40))
    file.write(b"\x00" * 24)

    entry_offsets = []
    for idx in range(len(entries)):
        offset = len(entries) * 4 + 40 + idx * 20
        entry_offsets.append(offset)
        file.write(struct.pack("<I", offset))

    for entry in entries:
        file.write(entry.resref.ljust(16, "\x00").encode("ascii"))
        file.write(struct.pack("<I", entry.strref))



"""
Read NDB files: NWScript compiled debug information.

The NDB format is used to store debugging information for
compiled scripts, such as variable and function names, line numbers, and
structure definitions.
"""

from enum import Enum
from dataclasses import dataclass
from typing import TextIO


class ScalarType(Enum):
    FLOAT = "f"
    INT = "i"
    VOID = "v"
    OBJECT = "o"
    STRING = "s"
    EFFECT = "e"


@dataclass(frozen=True)
class StructRef:
    id: int


@dataclass(frozen=True)
class Function:
    label: str
    b_start: int  # NB: includes NCS header
    b_end: int
    ret_type: ScalarType | StructRef
    args: list[ScalarType | StructRef]


@dataclass(frozen=True)
class Struct:
    label: str
    fields: list


@dataclass(frozen=True)
class Variable:
    label: str
    type: int
    b_start: int
    b_end: int
    stack_loc: int


@dataclass(frozen=True)
class Line:
    file_num: int
    line_num: int
    b_start: int
    b_end: int


@dataclass(frozen=True)
class Ndb:
    files: list
    structs: list
    functions: list
    variables: list
    lines: list

    def struct_by_id(self, struct_ref: StructRef) -> Struct:
        return self.structs[struct_ref.id]

    def function_by_name(self, name: str) -> Function:
        for f in self.functions:
            if f.label == name:
                return f
        raise KeyError(f"Function {name} not found")


def _parse_type(s):
    if s[0] == "t":
        return StructRef(int(s[1:]))
    return ScalarType(s[0])


def read(file: TextIO) -> Ndb:
    """
    Reads the given file object containing NDB data and returns an Ndb object.

    Args:
        file: A text mode file containing the NDB data to be parsed.

    Returns:
        Ndb: The parsed Ndb object.

    Raises:
        ValueError: If parsing fails.
    """

    files, structs, functions, variables, lines = [], [], [], [], []
    if file.readline().strip() != "NDB V1.0":
        raise ValueError("Invalid NDB file version.")

    counters = list(map(int, file.readline().strip().split(" ")))
    if len(counters) != 5:
        raise ValueError("Invalid NDB file header: Not enough counters.")

    while ln := file.readline().strip():
        if not ln or ln[0] == "#":
            continue

        s = ln.split(" ")

        if not s[0]:
            raise ValueError(f"Invalid line: {ln}")

        if s[0][0] in ["N", "n"]:
            if int(s[0][1:]) != len(files):
                raise ValueError(f"Expected file {len(files)}, got {s[0][1:]}")
            files.append(s[1])

        elif s[0] == "s":
            if len(s) != 3:
                raise ValueError(f"Invalid struct line: {ln}")
            structs.append(Struct(s[2], []))

        elif s[0] == "sf":
            if len(s) != 3 or not structs:
                raise ValueError(f"Invalid struct field line: {ln}")
            structs[-1].fields.append((s[2], _parse_type(s[1])))

        elif s[0] == "f":
            if len(s) != 6:
                raise ValueError(f"Invalid function line: {ln}")
            functions.append(
                Function(s[5], int(s[1], 16), int(s[2], 16), _parse_type(s[4]), [])
            )

        elif s[0] == "fp":
            if len(s) != 2 or not functions:
                raise ValueError(f"Invalid function parameter line: {ln}")
            functions[-1].args.append(_parse_type(s[1]))

        elif s[0] == "v":
            if len(s) != 6:
                raise ValueError(f"Invalid variable line: {ln}")
            variables.append(
                Variable(
                    s[5], _parse_type(s[4]), int(s[1], 16), int(s[2], 16), int(s[3], 16)
                )
            )

        elif s[0][0] == "l":
            lines.append(Line(int(s[0][1:]), int(s[1]), int(s[2], 16), int(s[3], 16)))

        else:
            raise ValueError(f"Unparsed line: {ln}")

    if len(files) != counters[0]:
        raise ValueError(f"Expected {counters[0]} files, got {len(files)}")
    if len(structs) != counters[1]:
        raise ValueError(f"Expected {counters[1]} structs, got {len(structs)}")
    if len(functions) != counters[2]:
        raise ValueError(f"Expected {counters[2]} functions, got {len(functions)}")
    if len(variables) != counters[3]:
        raise ValueError(f"Expected {counters[3]} variables, got {len(variables)}")
    if len(lines) != counters[4]:
        raise ValueError(f"Expected {counters[4]} lines, got {len(lines)}")

    return Ndb(
        files,
        structs,
        functions,
        variables,
        lines,
    )



"""
Transform GFF (Generic File Format) files from/to native python types.

Python representation is (Struct) and (List) objects, which can be nested.
They behave just like native python dictionaries and lists, but with some
additional methods and properties to make life easier.

Since GFF has strong typing beyond what python offers natively, the module
provides a number of custom types to represent the various field types that
can be found in a GFF file.

All field types are subclasses of the native python types, and are used to
enforce the GFF type system.

This module also provides helpers to transform NWN-style JSON data to/from
the native python type schema (eg. this data is returned by the game template
functions, and is also the serialisation format used by neverwinter.nim).
"""

from nwn.gff._reader import read
from nwn.gff._writer import write
from nwn.gff._types import (
    Byte,
    Char,
    Word,
    Short,
    Dword,
    Int,
    Dword64,
    Int64,
    Float,
    Double,
    CExoString,
    ResRef,
    CExoLocString,
    VOID,
    Struct,
    List,
    type_label_to_type,
)
from nwn.gff._json import struct_to_json, struct_from_json


__all__ = [
    "read",
    "write",
    "Byte",
    "Char",
    "Word",
    "Short",
    "Dword",
    "Int",
    "Dword64",
    "Int64",
    "Float",
    "Double",
    "CExoString",
    "ResRef",
    "CExoLocString",
    "VOID",
    "List",
    "Struct",
    "struct_to_json",
    "struct_from_json",
    "type_label_to_type",
]
